import unittest
from unittest.mock import Mock, patch

from parameterized import parameterized  # type: ignore

from jupyter_deploy.engine.outdefs import StrTemplateOutputDefinition
from jupyter_deploy.engine.terraform.tf_outdefs import (
    TerraformOutputDefinition,
    TerraformStrOutputDefinition,
    create_tf_output_definition,
    parse_output_cmd_result,
)
from jupyter_deploy.engine.terraform.tf_types import TerraformType


class TestTerraformOutputDefinition(unittest.TestCase):
    def test_instantiates_str_output_def(self) -> None:
        # Arrange
        output_name = "test_output"
        description = "Test description"
        value = "test_value"

        # Act
        output_def = TerraformStrOutputDefinition(output_name=output_name, description=description, value=value)
        template_def = output_def.to_template_definition()

        # Assert
        self.assertIsInstance(template_def, StrTemplateOutputDefinition)
        self.assertEqual(template_def.output_name, output_name)
        self.assertEqual(template_def.description, description)
        self.assertEqual(template_def.value, value)


class TestCreateTfOutputDefinition(unittest.TestCase):
    @parameterized.expand(
        [
            (
                "string_type",
                {"output_name": "str_var", "type": TerraformType.STRING},
                TerraformStrOutputDefinition,
            ),
        ]
    )
    def test_map_to_right_case(self, _name: str, parsed_config: dict, expected_class: type) -> None:
        # Act
        result = create_tf_output_definition(parsed_config)

        # Assert
        self.assertIsInstance(result, expected_class)
        self.assertEqual(result.output_name, parsed_config["output_name"])

    def test_raises_not_implemented_error_for_unknown_type(self) -> None:
        # Arrange
        parsed_config = {"output_name": "unknown_var", "type": "unknown_type"}

        # Act & Assert
        with self.assertRaises(NotImplementedError):
            create_tf_output_definition(parsed_config)


class TestParseOutputCmdResult(unittest.TestCase):
    @patch("json.loads")
    def test_parse_content_as_json(self, mock_json_loads: Mock) -> None:
        # Arrange
        content = '{"output1": {"type": "string", "value": "value1"}}'
        mock_json_loads.return_value = {"output1": {"type": "string", "value": "value1"}}

        # Act
        parse_output_cmd_result(content)

        # Assert
        mock_json_loads.assert_called_once_with(content)

    @patch("json.loads")
    def test_raises_runtime_error_if_parsed_json_not_a_dict(self, mock_json_loads: Mock) -> None:
        # Arrange
        content = "[1, 2, 3]"
        mock_json_loads.return_value = [1, 2, 3]

        # Act & Assert
        with self.assertRaises(RuntimeError):
            parse_output_cmd_result(content)

    @patch("jupyter_deploy.engine.terraform.tf_outdefs.create_tf_output_definition")
    @patch("json.loads")
    def test_parse_valid_output_def_and_add_name(self, mock_json_loads: Mock, mock_create_tf_output_def: Mock) -> None:
        # Arrange
        content = '{"output1": {"type": "string", "value": "value1"}}'
        mock_json_loads.return_value = {"output1": {"type": "string", "value": "value1"}}

        mock_output_def = Mock(spec=TerraformOutputDefinition)
        mock_create_tf_output_def.return_value = mock_output_def

        # Act
        result = parse_output_cmd_result(content)

        # Assert
        mock_create_tf_output_def.assert_called_once_with(
            {"type": "string", "value": "value1", "output_name": "output1"}
        )
        self.assertEqual(result, {"output1": mock_output_def})

    @patch("json.loads")
    def test_raises_runtime_error_if_name_not_str(self, mock_json_loads: Mock) -> None:
        # Arrange
        content = '{123: {"type": "string", "value": "value1"}}'
        mock_json_loads.return_value = {123: {"type": "string", "value": "value1"}}

        # Act & Assert
        with self.assertRaises(RuntimeError):
            parse_output_cmd_result(content)

    @patch("json.loads")
    def test_raises_runtime_error_if_config_not_dict(self, mock_json_loads: Mock) -> None:
        # Arrange
        content = '{"output1": "not a dict"}'
        mock_json_loads.return_value = {"output1": "not a dict"}

        # Act & Assert
        with self.assertRaises(RuntimeError):
            parse_output_cmd_result(content)



import inspect
import unittest
from pathlib import Path
from typing import Any
from unittest.mock import MagicMock, Mock, patch

import typer

from jupyter_deploy.cli.variables_decorator import with_project_variables
from jupyter_deploy.engine.vardefs import (
    BoolTemplateVariableDefinition,
    DictStrTemplateVariableDefinition,
    IntTemplateVariableDefinition,
    ListStrTemplateVariableDefinition,
    StrTemplateVariableDefinition,
)


class TestWithTemplateVariableDecorator(unittest.TestCase):
    @patch("jupyter_deploy.cli.variables_decorator.Path.cwd")
    @patch("jupyter_deploy.cli.variables_decorator.base_project_handler.retrieve_project_manifest_if_available")
    @patch("jupyter_deploy.cli.variables_decorator.variables_handler.VariablesHandler")
    def test_does_not_instantiate_handler_if_manifest_not_found(
        self, mock_handler_class: Mock, mock_retrieve_manifest: Mock, mock_cwd: Mock
    ) -> None:
        # Setup
        path = Path("/mock/cwd")
        mock_cwd.return_value = path
        mock_retrieve_manifest.return_value = None

        mock_handler = MagicMock()
        mock_handler_class.return_value = mock_handler

        # Define a test function
        def test_fn(arg1: str, variables: dict[str, Any] | None = None) -> None:
            pass

        # Apply decorator
        decorated_fn = with_project_variables()(test_fn)

        # Assert
        sig = inspect.signature(decorated_fn)
        self.assertIn("arg1", sig.parameters)
        self.assertNotIn("variables", sig.parameters)
        mock_handler_class.assert_not_called()
        mock_retrieve_manifest.assert_called_once_with(path)

    @patch("jupyter_deploy.cli.variables_decorator.Path.cwd")
    @patch("jupyter_deploy.cli.variables_decorator.base_project_handler.retrieve_project_manifest_if_available")
    @patch("jupyter_deploy.cli.variables_decorator.variables_handler.VariablesHandler")
    def test_returns_wrapper_without_variables_argument(
        self, mock_handler_class: Mock, mock_retrieve_manifest: Mock, mock_cwd: Mock
    ) -> None:
        # Setup
        path = Path("/mock/cwd")
        mock_cwd.return_value = path
        mock_manifest = Mock()
        mock_retrieve_manifest.return_value = mock_manifest

        mock_handler = MagicMock()
        mock_handler_class.return_value = mock_handler
        mock_handler.get_template_variables.return_value = {}

        # Define a test function
        def test_fn(arg1: str, variables: dict[str, Any] | None = None) -> None:
            pass

        # Apply decorator
        decorated_fn = with_project_variables()(test_fn)

        # Assert
        sig = inspect.signature(decorated_fn)
        self.assertIn("arg1", sig.parameters)
        self.assertNotIn("variables", sig.parameters)
        mock_handler_class.assert_called_once_with(project_path=path, project_manifest=mock_manifest)
        mock_retrieve_manifest.assert_called_once_with(path)

    @patch("jupyter_deploy.cli.variables_decorator.Path.cwd")
    @patch("jupyter_deploy.cli.variables_decorator.base_project_handler.retrieve_project_manifest_if_available")
    @patch("jupyter_deploy.cli.variables_decorator.variables_handler.VariablesHandler")
    def test_returns_wrapper_with_new_arguments_for_template_variables(
        self, mock_handler_class: Mock, mock_retrieve_manifest: Mock, mock_cwd: Mock
    ) -> None:
        # Setup
        mock_cwd.return_value = Path("/to/cur/dir")
        mock_retrieve_manifest.return_value = Mock()
        mock_handler = MagicMock()
        mock_handler_class.return_value = mock_handler
        mock_handler.get_template_variables.return_value = {
            "test_var": StrTemplateVariableDefinition(variable_name="test_var", description="Test variable")
        }

        # Define a test function
        def test_fn(arg1: str, variables: dict[str, Any] | None = None) -> None:
            pass

        # Apply decorator
        decorated_fn = with_project_variables()(test_fn)

        # Assert
        sig = inspect.signature(decorated_fn)
        self.assertIn("arg1", sig.parameters)
        self.assertIn("test_var", sig.parameters)
        self.assertNotIn("variables", sig.parameters)

    @patch("jupyter_deploy.cli.variables_decorator.Path.cwd")
    @patch("jupyter_deploy.cli.variables_decorator.base_project_handler.retrieve_project_manifest_if_available")
    @patch("jupyter_deploy.cli.variables_decorator.variables_handler.VariablesHandler")
    def test_template_variables_set_in_their_own_panel(
        self, mock_handler_class: Mock, mock_retrieve_manifest: Mock, mock_cwd: Mock
    ) -> None:
        # Setup
        mock_cwd.return_value = Path("/to/cur/dir")
        mock_retrieve_manifest.return_value = Mock()
        mock_handler = MagicMock()
        mock_handler_class.return_value = mock_handler
        mock_handler.get_template_variables.return_value = {
            "test_var": StrTemplateVariableDefinition(variable_name="test_var", description="Test variable")
        }

        # Define a test function
        def test_fn(arg1: str, variables: dict[str, Any] | None = None) -> None:
            pass

        # Apply decorator
        decorated_fn = with_project_variables()(test_fn)

        # Assert
        sig = inspect.signature(decorated_fn)
        param = sig.parameters["test_var"]
        # Check that the annotation includes typer.Option with rich_help_panel
        self.assertTrue(hasattr(param.annotation, "__metadata__"))
        option = param.annotation.__metadata__[0]
        self.assertIsInstance(option, typer.models.OptionInfo)
        self.assertEqual(option.rich_help_panel, "Template variables")

    @patch("jupyter_deploy.cli.variables_decorator.Path.cwd")
    @patch("jupyter_deploy.cli.variables_decorator.base_project_handler.retrieve_project_manifest_if_available")
    @patch("jupyter_deploy.cli.variables_decorator.variables_handler.VariablesHandler")
    def test_template_variables_have_default_set_to_none(
        self, mock_handler_class: Mock, mock_retrieve_manifest: Mock, mock_cwd: Mock
    ) -> None:
        # Setup
        mock_cwd.return_value = Path("/to/cur/dir")
        mock_retrieve_manifest.return_value = Mock()
        mock_handler = MagicMock()
        mock_handler_class.return_value = mock_handler
        mock_handler.get_template_variables.return_value = {
            "test_var": StrTemplateVariableDefinition(
                variable_name="test_var", description="Test variable", default="default_value"
            )
        }

        # Define a test function
        def test_fn(arg1: str, variables: dict[str, Any] | None = None) -> None:
            pass

        # Apply decorator
        decorated_fn = with_project_variables()(test_fn)

        # Assert
        sig = inspect.signature(decorated_fn)
        param = sig.parameters["test_var"]
        # Check that the default is None, not the variable's default value
        self.assertIsNone(param.default)

    @patch("jupyter_deploy.cli.variables_decorator.Path.cwd")
    @patch("jupyter_deploy.cli.variables_decorator.base_project_handler.retrieve_project_manifest_if_available")
    @patch("jupyter_deploy.cli.variables_decorator.variables_handler.VariablesHandler")
    def test_template_variables_have_type_set_correctly(
        self, mock_handler_class: Mock, mock_retrieve_manifest: Mock, mock_cwd: Mock
    ) -> None:
        # Setup
        mock_cwd.return_value = Path("/to/cur/dir")
        mock_retrieve_manifest.return_value = Mock()
        mock_handler = MagicMock()
        mock_handler_class.return_value = mock_handler
        mock_handler.get_template_variables.return_value = {
            "str_var": StrTemplateVariableDefinition(variable_name="str_var", description="String variable"),
            "int_var": IntTemplateVariableDefinition(variable_name="int_var", description="Integer variable"),
            "bool_var": BoolTemplateVariableDefinition(variable_name="bool_var", description="Boolean variable"),
            "list_str_var": ListStrTemplateVariableDefinition(
                variable_name="list_str_var", description="List variable"
            ),
            "map_str_var": DictStrTemplateVariableDefinition(variable_name="map_str_var", description="Map variable"),
        }

        # Define a test function
        def test_fn(arg1: str, variables: dict[str, Any] | None = None) -> None:
            pass

        # Apply decorator
        decorated_fn = with_project_variables()(test_fn)

        # Assert
        sig = inspect.signature(decorated_fn)

        # Check str_var type
        str_param = sig.parameters["str_var"]
        # Check that the parameter has a typer.Option in its metadata
        self.assertTrue(hasattr(str_param.annotation, "__metadata__"))
        # Check that the first argument of the Annotated type is str
        self.assertEqual(str_param.annotation.__args__[0], str)

        # Check int_var type
        int_param = sig.parameters["int_var"]
        self.assertTrue(hasattr(int_param.annotation, "__metadata__"))
        self.assertEqual(int_param.annotation.__args__[0], int)

        # Check bool_var type
        bool_param = sig.parameters["bool_var"]
        self.assertTrue(hasattr(bool_param.annotation, "__metadata__"))
        self.assertEqual(bool_param.annotation.__args__[0], bool)

        # Check list_str_var type
        list_str_param = sig.parameters["list_str_var"]
        self.assertTrue(hasattr(list_str_param.annotation, "__metadata__"))
        self.assertEqual(list_str_param.annotation.__args__[0], list[str])

        # Check map_str_var type
        map_str_param = sig.parameters["map_str_var"]
        self.assertTrue(hasattr(map_str_param.annotation, "__metadata__"))
        self.assertEqual(map_str_param.annotation.__args__[0], list[str])

    @patch("jupyter_deploy.cli.variables_decorator.Path.cwd")
    @patch("jupyter_deploy.cli.variables_decorator.base_project_handler.retrieve_project_manifest_if_available")
    @patch("jupyter_deploy.cli.variables_decorator.variables_handler.VariablesHandler")
    def test_template_variables_are_shown_with_help(
        self, mock_handler_class: Mock, mock_retrieve_manifest: Mock, mock_cwd: Mock
    ) -> None:
        # Setup
        mock_cwd.return_value = Path("/to/cur/dir")
        mock_retrieve_manifest.return_value = Mock()
        mock_handler = MagicMock()
        mock_handler_class.return_value = mock_handler
        mock_handler.get_template_variables.return_value = {
            "test_var": StrTemplateVariableDefinition(variable_name="test_var", description="Test variable description")
        }

        # Define a test function
        def test_fn(arg1: str, variables: dict[str, Any] | None = None) -> None:
            pass

        # Apply decorator
        decorated_fn = with_project_variables()(test_fn)

        # Assert
        sig = inspect.signature(decorated_fn)
        param = sig.parameters["test_var"]
        option = param.annotation.__metadata__[0]
        self.assertEqual(option.help, "Test variable description")

    @patch("jupyter_deploy.cli.variables_decorator.Path.cwd")
    @patch("jupyter_deploy.cli.variables_decorator.base_project_handler.retrieve_project_manifest_if_available")
    @patch("jupyter_deploy.cli.variables_decorator.variables_handler.VariablesHandler")
    def test_default_reflected_as_preset_in_variable_help(
        self, mock_handler_class: Mock, mock_retrieve_manifest: Mock, mock_cwd: Mock
    ) -> None:
        # Setup
        mock_cwd.return_value = Path("/to/cur/dir")
        mock_retrieve_manifest.return_value = Mock()
        mock_handler = MagicMock()
        mock_handler_class.return_value = mock_handler
        mock_handler.get_template_variables.return_value = {
            "test_var": StrTemplateVariableDefinition(
                variable_name="test_var", description="Test variable", default="default_value"
            )
        }

        # Define a test function
        def test_fn(arg1: str, variables: dict[str, Any] | None = None) -> None:
            pass

        # Apply decorator
        decorated_fn = with_project_variables()(test_fn)

        # Assert
        sig = inspect.signature(decorated_fn)
        param = sig.parameters["test_var"]
        option = param.annotation.__metadata__[0]
        self.assertIn("[preset: default_value]", option.help)

    @patch("jupyter_deploy.cli.variables_decorator.Path.cwd")
    @patch("jupyter_deploy.cli.variables_decorator.base_project_handler.retrieve_project_manifest_if_available")
    @patch("jupyter_deploy.cli.variables_decorator.variables_handler.VariablesHandler")
    def test_default_not_reflected_in_list_and_map_variables_help(
        self, mock_handler_class: Mock, mock_retrieve_manifest: Mock, mock_cwd: Mock
    ) -> None:
        # Setup
        mock_cwd.return_value = Path("/to/cur/dir")
        mock_retrieve_manifest.return_value = Mock()
        mock_handler = MagicMock()
        mock_handler_class.return_value = mock_handler
        mock_handler.get_template_variables.return_value = {
            "list_var": ListStrTemplateVariableDefinition(
                variable_name="list_var", description="Test list", default=["a", "b"]
            ),
            "map_var": DictStrTemplateVariableDefinition(variable_name="map_var", description="Test map", default={}),
        }

        # Define a test function
        def test_fn(arg1: str, variables: dict[str, Any] | None = None) -> None:
            pass

        # Apply decorator
        decorated_fn = with_project_variables()(test_fn)

        # Assert
        sig = inspect.signature(decorated_fn)
        list_param = sig.parameters["list_var"]
        list_option = list_param.annotation.__metadata__[0]
        self.assertNotIn("[preset: ", list_option.help)
        self.assertIn("--list-var", list_option.help)

        map_param = sig.parameters["map_var"]
        map_option = map_param.annotation.__metadata__[0]
        self.assertNotIn("[preset: ", map_option.help)
        self.assertIn("--map-var", map_option.help)

    @patch("jupyter_deploy.cli.variables_decorator.Path.cwd")
    @patch("jupyter_deploy.cli.variables_decorator.base_project_handler.retrieve_project_manifest_if_available")
    @patch("jupyter_deploy.cli.variables_decorator.variables_handler.VariablesHandler")
    def test_map_str_variables_are_passed_a_callback(
        self, mock_handler_class: Mock, mock_retrieve_manifest: Mock, mock_cwd: Mock
    ) -> None:
        # Setup
        mock_cwd.return_value = Path("/to/cur/dir")
        mock_retrieve_manifest.return_value = Mock()
        mock_handler = MagicMock()
        mock_handler_class.return_value = mock_handler
        mock_handler.get_template_variables.return_value = {
            "test_var": DictStrTemplateVariableDefinition(
                variable_name="test_var", description="Test variable description"
            )
        }

        # Define a test function
        def test_fn(arg1: str, variables: dict[str, Any] | None = None) -> None:
            pass

        # Apply decorator
        decorated_fn = with_project_variables()(test_fn)

        # Assert
        sig = inspect.signature(decorated_fn)
        param = sig.parameters["test_var"]
        option = param.annotation.__metadata__[0]
        self.assertTrue(hasattr(option, "callback"))
        self.assertIsNotNone(option.callback)

    @patch("jupyter_deploy.cli.variables_decorator.Path.cwd")
    @patch("jupyter_deploy.cli.variables_decorator.base_project_handler.retrieve_project_manifest_if_available")
    @patch("jupyter_deploy.cli.variables_decorator.variables_handler.VariablesHandler")
    def test_pass_vars_assigned_a_value_by_caller_to_inner_method_via_variables_attribute(
        self, mock_handler_class: Mock, mock_retrieve_manifest: Mock, mock_cwd: Mock
    ) -> None:
        # Setup
        mock_cwd.return_value = Path("/to/cur/dir")
        mock_retrieve_manifest.return_value = Mock()
        mock_handler = MagicMock()
        mock_handler_class.return_value = mock_handler
        mock_handler.get_template_variables.return_value = {
            "test_var": StrTemplateVariableDefinition(variable_name="test_var", description="Test variable")
        }

        # Define a test function with a mock to capture the arguments
        mock_fn = Mock()

        def test_fn(arg1: str, variables: dict[str, Any] | None = None) -> None:
            mock_fn(arg1=arg1, variables=variables)

        # Apply decorator
        decorated_fn = with_project_variables()(test_fn)

        # Call the decorated function with a value for test_var
        decorated_fn(arg1="test", test_var="assigned_value")

        # Assert
        mock_fn.assert_called_once()
        call_args = mock_fn.call_args[1]
        self.assertEqual(call_args["arg1"], "test")
        self.assertIn("test_var", call_args["variables"])
        self.assertTrue(isinstance(call_args["variables"]["test_var"], StrTemplateVariableDefinition))
        self.assertEqual(call_args["variables"]["test_var"].assigned_value, "assigned_value")

    @patch("jupyter_deploy.cli.variables_decorator.Path.cwd")
    @patch("jupyter_deploy.cli.variables_decorator.base_project_handler.retrieve_project_manifest_if_available")
    @patch("jupyter_deploy.cli.variables_decorator.variables_handler.VariablesHandler")
    def test_vars_not_assigned_a_value_by_caller_not_passed_to_inner_method(
        self, mock_handler_class: Mock, mock_retrieve_manifest: Mock, mock_cwd: Mock
    ) -> None:
        # Setup
        mock_cwd.return_value = Path("/to/cur/dir")
        mock_retrieve_manifest.return_value = Mock()
        mock_handler = MagicMock()
        mock_handler_class.return_value = mock_handler
        mock_handler.get_template_variables.return_value = {
            "test_str_var": StrTemplateVariableDefinition(
                variable_name="test_str_var", description="Test str variable"
            ),
            "test_int_var": StrTemplateVariableDefinition(
                variable_name="test_int_var", description="Test int variable"
            ),
            "another_var": StrTemplateVariableDefinition(variable_name="another_var", description="Another variable"),
        }

        # Define a test function with a mock to capture the arguments
        mock_fn = Mock()

        def test_fn(arg1: str, variables: dict[str, Any] | None = None) -> None:
            mock_fn(arg1=arg1, variables=variables)

        # Apply decorator
        decorated_fn = with_project_variables()(test_fn)

        # Call the decorated function with a value for only test_str_var and test_int_var
        decorated_fn(arg1="test", test_str_var="assigned_value", test_int_var=30, another_var=None)

        # Assert
        mock_fn.assert_called_once()
        call_args = mock_fn.call_args[1]
        self.assertEqual(call_args["arg1"], "test")
        self.assertIn("test_str_var", call_args["variables"])
        self.assertIn("test_int_var", call_args["variables"])
        self.assertNotIn("another_var", call_args["variables"])

    @patch("jupyter_deploy.cli.variables_decorator.Path.cwd")
    @patch("jupyter_deploy.cli.variables_decorator.base_project_handler.retrieve_project_manifest_if_available")
    @patch("jupyter_deploy.cli.variables_decorator.variables_handler.VariablesHandler")
    def test_dict_vars_translated_to_inner_method_before_passing_to_variables(
        self, mock_handler_class: Mock, mock_retrieve_manifest: Mock, mock_cwd: Mock
    ) -> None:
        # Setup
        mock_cwd.return_value = Path("/to/cur/dir")
        mock_retrieve_manifest.return_value = Mock()
        mock_handler = MagicMock()
        mock_handler_class.return_value = mock_handler
        mock_handler.get_template_variables.return_value = {
            "test_var": DictStrTemplateVariableDefinition(variable_name="test_var", description="Test variable")
        }

        # Define a test function with a mock to capture the arguments
        mock_fn = Mock()

        def test_fn(arg1: str, variables: dict[str, Any] | None = None) -> None:
            mock_fn(arg1=arg1, variables=variables)

        # Apply decorator
        decorated_fn = with_project_variables()(test_fn)

        # Call the decorated function with a value for test_var
        decorated_fn(arg1="test", test_var=["key1=val1", "key2=val2"])

        # Assert
        mock_fn.assert_called_once()
        call_args = mock_fn.call_args[1]
        self.assertEqual(call_args["arg1"], "test")
        self.assertIn("test_var", call_args["variables"])
        self.assertTrue(isinstance(call_args["variables"]["test_var"], DictStrTemplateVariableDefinition))
        self.assertEqual(call_args["variables"]["test_var"].assigned_value, {"key1": "val1", "key2": "val2"})



from typing import Generic, TypeVar

from pydantic import BaseModel, ConfigDict

from jupyter_deploy import str_utils

S = TypeVar("S")


class TemplateOutputDefinition(BaseModel, Generic[S]):
    """Wrapper class for template-outputable values."""

    model_config = ConfigDict(extra="allow")
    output_name: str
    description: str = ""
    sensitive: bool = False
    value: S | None = None

    def get_cli_description(self) -> str:
        """Return a one-liner description with value information for the CLI attribute."""
        header = str_utils.get_trimmed_header(self.description)
        value_marker = f"value: {self.value}" if self.value is not None else ""
        separator = " " if len(header) > 0 and len(value_marker) else ""
        return f"{header}{separator}{value_marker}"


class StrTemplateOutputDefinition(TemplateOutputDefinition[str]):
    """Wrapper class for template-outputable string value."""

    pass


class ListStrTemplateOutputDefinition(TemplateOutputDefinition[list[str]]):
    """Wrapper class for teemplate-outputable list of string value."""

    pass


TOD = TypeVar("TOD", bound=TemplateOutputDefinition)


def require_output_def(
    output_defs: dict[str, TemplateOutputDefinition], output_name: str, output_type: type[TOD]
) -> TOD:
    """Returns the resolved output of the expected type from the ouput_defs dict.

    Args:
        output_defs: A dictionary of outputs.
        output_name: The name of the output to retrieve.
        output_type: The expected type of the output def, a subclass of TemplateOutputDefinition.

    Raises:
        KeyError: If output_name is not in resolved_args.
        TypeError: If the argument is not of the expected type.
    """
    if output_name not in output_defs:
        raise KeyError(f"Required output '{output_name}' not found in output definitions")

    output_def = output_defs[output_name]

    if not isinstance(output_def, output_type):
        raise TypeError(
            f"Expected output '{output_def}' to be of type {output_type.__name__}, got {type(output_def).__name__}"
        )

    return output_def



import unittest
from pathlib import Path
from unittest.mock import Mock, patch

from jupyter_deploy.engine.engine_open import EngineOpenHandler
from jupyter_deploy.engine.outdefs import StrTemplateOutputDefinition


class TestEngineOpen(unittest.TestCase):
    def get_mock_outputs_handler_and_fns(self) -> tuple[Mock, dict[str, Mock]]:
        """Return a mocked outputs handler."""
        mock_handler = Mock()
        mock_get_declared_output_def = Mock()
        mock_handler.get_declared_output_def = mock_get_declared_output_def

        mock_get_declared_output_def.return_value = StrTemplateOutputDefinition(
            output_name="notebook_url", value="https://notebook.my.domain"
        )

        return mock_handler, {"get_declared_output_def": mock_get_declared_output_def}

    def get_mock_console_and_fns(self) -> tuple[Mock, dict[str, Mock]]:
        """Return a mocked rich console instance."""
        mock_console = Mock()
        mock_print = Mock()
        mock_console.print = mock_print
        return mock_console, {"print": mock_print}

    def test_init(self) -> None:
        mock_outputs_handler, mock_outputs_handler_fns = self.get_mock_outputs_handler_and_fns()
        mock_manifest = Mock()
        handler = EngineOpenHandler(
            project_path=Path("/project/path"),
            project_manifest=mock_manifest,
            output_handler=mock_outputs_handler,
        )
        self.assertEqual(handler.project_path, Path("/project/path"))
        self.assertEqual(handler.project_manifest, mock_manifest)
        self.assertEqual(handler.output_handler, mock_outputs_handler)
        mock_outputs_handler_fns["get_declared_output_def"].assert_not_called()

    def test_get_url_happy_case(self) -> None:
        # Setup
        mock_outputs_handler, mock_outputs_handler_fns = self.get_mock_outputs_handler_and_fns()
        mock_manifest = Mock()
        handler = EngineOpenHandler(
            project_path=Path("/project/path"),
            project_manifest=mock_manifest,
            output_handler=mock_outputs_handler,
        )

        # Act
        url = handler.get_url()

        # Verify
        self.assertEqual(url, "https://notebook.my.domain")
        mock_outputs_handler_fns["get_declared_output_def"].assert_called_once_with(
            "open_url", StrTemplateOutputDefinition
        )

    @patch("rich.console.Console")
    def test_get_url_not_implemented_case(self, mock_console_cls: Mock) -> None:
        # Setup
        mock_console, mock_console_fns = self.get_mock_console_and_fns()
        mock_console_cls.return_value = mock_console

        mock_outputs_handler, mock_outputs_handler_fns = self.get_mock_outputs_handler_and_fns()
        mock_outputs_handler_fns["get_declared_output_def"].side_effect = NotImplementedError("open_url not declared")
        mock_manifest = Mock()
        handler = EngineOpenHandler(
            project_path=Path("/project/path"),
            project_manifest=mock_manifest,
            output_handler=mock_outputs_handler,
        )

        # Act
        url = handler.get_url()

        # Verify
        self.assertEqual(url, "")
        mock_outputs_handler_fns["get_declared_output_def"].assert_called_once()
        mock_console_fns["print"].assert_called_once()

    @patch("rich.console.Console")
    def test_get_url_value_error_case(self, mock_console_cls: Mock) -> None:
        # Setup
        mock_console, mock_console_fns = self.get_mock_console_and_fns()
        mock_console_cls.return_value = mock_console

        mock_outputs_handler, mock_outputs_handler_fns = self.get_mock_outputs_handler_and_fns()
        mock_outputs_handler_fns["get_declared_output_def"].side_effect = ValueError("open_url not a template output")
        mock_manifest = Mock()
        handler = EngineOpenHandler(
            project_path=Path("/project/path"),
            project_manifest=mock_manifest,
            output_handler=mock_outputs_handler,
        )

        # Act
        url = handler.get_url()

        # Verify
        self.assertEqual(url, "")
        mock_outputs_handler_fns["get_declared_output_def"].assert_called_once()
        mock_console_fns["print"].assert_called_once()

    @patch("rich.console.Console")
    def test_get_url_wrong_type_case(self, mock_console_cls: Mock) -> None:
        # Setup
        mock_console, mock_console_fns = self.get_mock_console_and_fns()
        mock_console_cls.return_value = mock_console

        mock_outputs_handler, mock_outputs_handler_fns = self.get_mock_outputs_handler_and_fns()
        mock_outputs_handler_fns["get_declared_output_def"].side_effect = TypeError("open_url not a str")
        mock_manifest = Mock()
        handler = EngineOpenHandler(
            project_path=Path("/project/path"),
            project_manifest=mock_manifest,
            output_handler=mock_outputs_handler,
        )

        # Act
        url = handler.get_url()

        # Verify
        self.assertEqual(url, "")
        mock_outputs_handler_fns["get_declared_output_def"].assert_called_once()
        mock_console_fns["print"].assert_called_once()

    @patch("rich.console.Console")
    def test_get_url_none_value_case(self, mock_console_cls: Mock) -> None:
        # Setup
        mock_console, mock_console_fns = self.get_mock_console_and_fns()
        mock_console_cls.return_value = mock_console

        mock_outputs_handler, mock_outputs_handler_fns = self.get_mock_outputs_handler_and_fns()
        mock_outputs_handler_fns["get_declared_output_def"].return_value = StrTemplateOutputDefinition(
            output_name="notebook_url"
        )
        mock_manifest = Mock()
        handler = EngineOpenHandler(
            project_path=Path("/project/path"),
            project_manifest=mock_manifest,
            output_handler=mock_outputs_handler,
        )

        # Act
        url = handler.get_url()

        # Verify
        self.assertEqual(url, "")
        mock_outputs_handler_fns["get_declared_output_def"].assert_called_once()
        mock_console_fns["print"].assert_called_once()



import unittest
from pathlib import Path
from typing import Any
from unittest.mock import Mock, patch

from pydantic import ValidationError

from jupyter_deploy import constants
from jupyter_deploy.engine.engine_variables import EngineVariablesHandler
from jupyter_deploy.engine.vardefs import TemplateVariableDefinition
from jupyter_deploy.handlers import base_project_handler
from jupyter_deploy.variables_config import (
    VARIABLES_CONFIG_V1_COMMENTS,
    VARIABLES_CONFIG_V1_KEYS_ORDER,
    JupyterDeployVariablesConfig,
    JupyterDeployVariablesConfigV1,
)


# Create a dummy implementation of EngineVariablesHandler for testing
class DummyVariablesHandler(EngineVariablesHandler):
    def is_template_directory(self) -> bool:
        return True

    def get_template_variables(self) -> dict[str, TemplateVariableDefinition]:
        variable1 = Mock(spec=TemplateVariableDefinition[str])
        variable1.has_default = True
        variable1.sensitive = False
        variable1.default = "default1"

        variable2 = Mock(spec=TemplateVariableDefinition[str])
        variable2.has_default = False
        variable2.sensitive = False
        variable2.default = None

        variable3 = Mock(spec=TemplateVariableDefinition[int])
        variable3.has_default = False
        variable3.sensitive = True
        variable3.default = None

        return {
            "var1": variable1,
            "var2": variable2,
            "var3": variable3,
        }

    def update_variable_records(self, varvalues: dict[str, Any], sensitive: bool = False) -> None:
        # This would normally update the engine-specific files
        pass


class TestVariablesConfigProperty(unittest.TestCase):
    def test_variables_config_not_defined_on_instantiation(self) -> None:
        # Verify that _variables_config is None after initialization
        project_path = Path("/mock/project")
        manifest = Mock()
        handler = DummyVariablesHandler(project_path=project_path, project_manifest=manifest)

        self.assertIsNone(handler._variables_config)

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_variables_config")
    def test_variables_config_read_config_on_access(self, mock_retrieve: Mock) -> None:
        # Setup
        project_path = Path("/mock/project")
        manifest = Mock()
        handler = DummyVariablesHandler(project_path=project_path, project_manifest=manifest)

        mock_config = Mock(spec=JupyterDeployVariablesConfig)
        mock_retrieve.return_value = mock_config

        # Access the property
        result = handler.variables_config

        # Verify that the config was retrieved and cached
        self.assertEqual(result, mock_config)
        self.assertEqual(handler._variables_config, mock_config)
        mock_retrieve.assert_called_once_with(project_path / constants.VARIABLES_FILENAME)

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_variables_config")
    def test_multiple_access_to_variables_config_reads_only_once(self, mock_retrieve: Mock) -> None:
        # Setup
        project_path = Path("/mock/project")
        manifest = Mock()
        handler = DummyVariablesHandler(project_path=project_path, project_manifest=manifest)

        mock_config = Mock(spec=JupyterDeployVariablesConfig)
        mock_retrieve.return_value = mock_config

        # Access the property multiple times
        result1 = handler.variables_config
        result2 = handler.variables_config

        # Verify that retrieve was called only once and both results match
        self.assertEqual(result1, mock_config)
        self.assertEqual(result2, mock_config)
        mock_retrieve.assert_called_once()

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_variables_config")
    def test_falls_back_to_empty_config_on_filenotfound_error(self, mock_retrieve: Mock) -> None:
        # Setup
        project_path = Path("/mock/project")
        manifest = Mock()
        handler = DummyVariablesHandler(project_path=project_path, project_manifest=manifest)

        # Mock retrieve to raise FileNotFoundError
        mock_retrieve.side_effect = FileNotFoundError("File not found")

        # Access the property
        result = handler.variables_config

        # Verify that a reset config was returned
        self.assertIsInstance(result, JupyterDeployVariablesConfigV1)
        mock_retrieve.assert_called_once()

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_variables_config")
    def test_falls_back_to_empty_config_on_validation_error(self, mock_retrieve: Mock) -> None:
        # Setup
        project_path = Path("/mock/project")
        manifest = Mock()
        handler = DummyVariablesHandler(project_path=project_path, project_manifest=manifest)

        # Mock retrieve to raise ValidationError
        mock_retrieve.side_effect = ValidationError.from_exception_data("Validation error", [])

        # Access the property
        result = handler.variables_config

        # Verify that a reset config was returned
        self.assertIsInstance(result, JupyterDeployVariablesConfigV1)
        mock_retrieve.assert_called_once()

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_variables_config")
    def test_falls_back_to_empty_config_on_notadict_error(self, mock_retrieve: Mock) -> None:
        # Setup
        project_path = Path("/mock/project")
        manifest = Mock()
        handler = DummyVariablesHandler(project_path=project_path, project_manifest=manifest)

        # Mock retrieve to raise NotADictError
        mock_retrieve.side_effect = base_project_handler.NotADictError("Invalid variables config")

        # Access the property
        result = handler.variables_config

        # Verify that a reset config was returned
        self.assertIsInstance(result, JupyterDeployVariablesConfigV1)
        mock_retrieve.assert_called_once()


class TestSyncEngineVarfilesWithProjectVariablesConfig(unittest.TestCase):
    def test_combines_required_and_overrides(self) -> None:
        # Setup
        project_path = Path("/mock/project")
        manifest = Mock()
        handler = DummyVariablesHandler(project_path=project_path, project_manifest=manifest)

        # Mock the handler's update_variable_records method
        with patch.object(handler, "update_variable_records") as mock_update_records:
            # Create a mock variables_config with required and overrides values
            mock_config = Mock(spec=JupyterDeployVariablesConfig)
            mock_config.required = {"var1": "value1", "var2": "value2"}
            mock_config.required_sensitive = {}
            mock_config.overrides = {"var3": "value3", "var4": "value4"}

            # Patch the variables_config property to return our mock
            handler._variables_config = mock_config

            # Execute
            handler.sync_engine_varfiles_with_project_variables_config()

            # Should combine required and overrides into one dictionary for non-sensitive variables
            expected_combined = {"var1": "value1", "var2": "value2", "var3": "value3", "var4": "value4"}
            mock_update_records.assert_any_call(expected_combined)

    def test_skips_none_values(self) -> None:
        # Setup
        project_path = Path("/mock/project")
        manifest = Mock()
        handler = DummyVariablesHandler(project_path=project_path, project_manifest=manifest)

        # Mock the handler's update_variable_records method
        with patch.object(handler, "update_variable_records") as mock_update_records:
            # Create a mock variables_config with required, sensitive, and overrides with some None values
            mock_config = Mock(spec=JupyterDeployVariablesConfig)
            mock_config.required = {"var1": "value1", "var2": None}
            mock_config.required_sensitive = {"var3": "value3", "var4": None}
            mock_config.overrides = {"var5": "value5", "var6": None}

            # Patch the variables_config property to return our mock
            handler._variables_config = mock_config

            # Execute
            handler.sync_engine_varfiles_with_project_variables_config()

            # Verify - None values should be skipped
            mock_update_records.assert_any_call({"var1": "value1", "var5": "value5"})
            mock_update_records.assert_any_call({"var3": "value3"}, sensitive=True)

    def test_calls_child_methods_on_variables_and_secrets(self) -> None:
        # Setup
        project_path = Path("/mock/project")
        manifest = Mock()
        handler = DummyVariablesHandler(project_path=project_path, project_manifest=manifest)

        # Mock the handler's update_variable_records method
        with patch.object(handler, "update_variable_records") as mock_update_records:
            # Create a mock variables_config with both regular and sensitive variables
            mock_config = Mock(spec=JupyterDeployVariablesConfig)
            mock_config.required = {"var1": "value1"}
            mock_config.required_sensitive = {"var2": "value2"}
            mock_config.overrides = {"var3": "value3"}

            # Patch the variables_config property to return our mock
            handler._variables_config = mock_config

            # Execute
            handler.sync_engine_varfiles_with_project_variables_config()

            # Verify that update_variable_records was called twice with correct arguments
            self.assertEqual(mock_update_records.call_count, 2)
            mock_update_records.assert_any_call({"var1": "value1", "var3": "value3"})
            mock_update_records.assert_any_call({"var2": "value2"}, sensitive=True)

    def test_raises_if_child_methods_raises(self) -> None:
        # Setup
        project_path = Path("/mock/project")
        manifest = Mock()
        handler = DummyVariablesHandler(project_path=project_path, project_manifest=manifest)

        # Mock the handler's update_variable_records method to raise an exception
        with patch.object(handler, "update_variable_records", side_effect=ValueError("Test error")):
            # Create a mock variables_config
            mock_config = Mock(spec=JupyterDeployVariablesConfig)
            mock_config.required = {"var1": "value1"}
            mock_config.required_sensitive = {}
            mock_config.overrides = {}

            # Patch the variables_config property to return our mock
            handler._variables_config = mock_config

            # Execute and verify that the exception propagates
            with self.assertRaises(ValueError):
                handler.sync_engine_varfiles_with_project_variables_config()


class TestGetVariableNamesAssignedInConfig(unittest.TestCase):
    def test_returns_combined_non_none_required_required_sensitive_and_overrides(self) -> None:
        # Setup
        project_path = Path("/mock/project")
        manifest = Mock()
        handler = DummyVariablesHandler(project_path=project_path, project_manifest=manifest)

        # Create a mock variables_config with a mix of None and non-None values
        mock_config = Mock(spec=JupyterDeployVariablesConfig)
        mock_config.required = {"var1": "value1", "var2": None}
        mock_config.required_sensitive = {"var3": "value3", "var4": None}
        mock_config.overrides = {"var5": "value5", "var6": None}

        # Patch the variables_config property to return our mock
        handler._variables_config = mock_config

        # Execute
        result = handler.get_variable_names_assigned_in_config()

        # Verify - only non-None values should be included
        self.assertEqual(len(result), 3)  # Only 3 non-None values
        self.assertIn("var1", result)  # From required
        self.assertIn("var3", result)  # From required_sensitive
        self.assertIn("var5", result)  # From overrides

        # Verify - None values should NOT be included
        self.assertNotIn("var2", result)  # None in required
        self.assertNotIn("var4", result)  # None in required_sensitive
        self.assertNotIn("var6", result)  # None in overrides

    def test_handles_empty_values_in_variables_config(self) -> None:
        # Setup
        project_path = Path("/mock/project")
        manifest = Mock()
        handler = DummyVariablesHandler(project_path=project_path, project_manifest=manifest)

        # Create a mock variables_config with empty dictionaries
        mock_config = Mock(spec=JupyterDeployVariablesConfig)
        mock_config.required = {}
        mock_config.required_sensitive = {}
        mock_config.overrides = {}

        # Patch the variables_config property to return our mock
        handler._variables_config = mock_config

        # Execute
        result = handler.get_variable_names_assigned_in_config()

        # Verify - should return an empty list
        self.assertEqual(result, [])

    def test_none_values_are_filtered_out(self) -> None:
        # Setup
        project_path = Path("/mock/project")
        manifest = Mock()
        handler = DummyVariablesHandler(project_path=project_path, project_manifest=manifest)

        # Create a mock variables_config with all None values
        mock_config = Mock(spec=JupyterDeployVariablesConfig)
        mock_config.required = {"var1": None, "var2": None}
        mock_config.required_sensitive = {"var3": None, "var4": None}
        mock_config.overrides = {"var5": None, "var6": None}

        # Patch the variables_config property to return our mock
        handler._variables_config = mock_config

        # Execute
        result = handler.get_variable_names_assigned_in_config()

        # Verify - should return an empty list since all values are None
        self.assertEqual(result, [])

    def test_defaults_are_ignored_even_when_not_none(self) -> None:
        # Setup
        project_path = Path("/mock/project")
        manifest = Mock()
        handler = DummyVariablesHandler(project_path=project_path, project_manifest=manifest)

        # Create a mock variables_config with defaults that have non-None values
        mock_config = Mock(spec=JupyterDeployVariablesConfig)
        mock_config.required = {"var1": None}
        mock_config.required_sensitive = {"var2": None}
        mock_config.overrides = {"var3": "value3"}
        mock_config.defaults = {"var4": "default4", "var5": "default5"}

        # Patch the variables_config property to return our mock
        handler._variables_config = mock_config

        # Execute
        result = handler.get_variable_names_assigned_in_config()

        # Verify - only non-None from required, required_sensitive and overrides should be included
        self.assertEqual(len(result), 1)  # Only 1 non-None value from the tracked sections
        self.assertIn("var3", result)  # From overrides

        # Verify - default values should be completely ignored, even though they're non-None
        self.assertNotIn("var4", result)
        self.assertNotIn("var5", result)


class TestSyncProjectVariablesConfig(unittest.TestCase):
    @patch("jupyter_deploy.fs_utils.write_yaml_file_with_comments")
    def test_handles_required_sensitive_overrides_and_defaults(self, mock_write: Mock) -> None:
        # Setup
        project_path = Path("/mock/project")
        manifest = Mock()
        handler = DummyVariablesHandler(project_path=project_path, project_manifest=manifest)

        # Create a mock variables_config
        mock_config = Mock(spec=JupyterDeployVariablesConfig)
        mock_config.required = {"var1": "value1"}
        mock_config.required_sensitive = {"var2": "value2"}
        mock_config.overrides = {}
        mock_config.defaults = {"var3": "value3", "var4": "value4", "var5": "value5"}

        # Create expected model_dump output
        expected_model_dump = {
            "schema_version": 1,
            "required": {"var1": "new1"},
            "required_sensitive": {"var2": "new2"},
            "overrides": {"var3": "new3", "var5": "new5"},
            "defaults": {"var3": "value3", "var4": "value4", "var5": "value5"},
        }

        # Create mock for new config
        mock_new_config = Mock()
        mock_new_config.model_dump.return_value = expected_model_dump

        # Patch the variables_config property to return our mock
        handler._variables_config = mock_config

        # Patch write_yaml_file_with_comments and JupyterDeployVariablesConfigV1
        # Execute
        updated_values = {"var1": "new1", "var2": "new2", "var3": "new3", "var5": "new5"}
        handler.sync_project_variables_config(updated_values)

        # Verify that the config was updated with the new values
        model_dump = mock_write.call_args[0][1]
        self.assertEqual(model_dump, expected_model_dump)

    @patch("jupyter_deploy.fs_utils.write_yaml_file_with_comments")
    def test_calls_write_with_comments_and_key_order(self, mock_write: Mock) -> None:
        # Setup
        project_path = Path("/mock/project")
        manifest = Mock()
        handler = DummyVariablesHandler(project_path=project_path, project_manifest=manifest)

        # Create a mock variables_config
        mock_config = Mock(spec=JupyterDeployVariablesConfig)
        mock_config.required = {}
        mock_config.required_sensitive = {}
        mock_config.overrides = {}
        mock_config.defaults = {"var1": "value1"}

        # Mock JupyterDeployVariablesConfigV1 to return a mock with model_dump method
        mock_new_config = Mock()
        mock_new_config.model_dump.return_value = {
            "schema_version": 1,
            "required": {},
            "required_sensitive": {},
            "overrides": {},
            "defaults": {"var1": "value1"},
        }

        # Patch the variables_config property to return our mock
        handler._variables_config = mock_config

        # Execute
        handler.sync_project_variables_config({"var1": "value1"})

        # Verify that write_yaml_file_with_comments was called with the correct arguments
        mock_write.assert_called_once()
        self.assertEqual(mock_write.call_args[0][0], project_path / constants.VARIABLES_FILENAME)
        self.assertEqual(mock_write.call_args[1]["key_order"], VARIABLES_CONFIG_V1_KEYS_ORDER)
        self.assertEqual(mock_write.call_args[1]["comments"], VARIABLES_CONFIG_V1_COMMENTS)

    @patch("jupyter_deploy.fs_utils.write_yaml_file_with_comments")
    def test_raises_when_write_method_raises(self, mock_write: Mock) -> None:
        # Setup
        project_path = Path("/mock/project")
        manifest = Mock()
        handler = DummyVariablesHandler(project_path=project_path, project_manifest=manifest)

        # Create a mock variables_config
        mock_config = Mock(spec=JupyterDeployVariablesConfig)
        mock_config.required = {}
        mock_config.required_sensitive = {}
        mock_config.overrides = {}
        mock_config.defaults = {"var1": "value1"}

        # Mock JupyterDeployVariablesConfigV1 to return a mock with model_dump method
        mock_new_config = Mock()
        mock_new_config.model_dump.return_value = {
            "schema_version": 1,
            "required": {},
            "required_sensitive": {},
            "overrides": {},
            "defaults": {"var1": "value1"},
        }

        # Patch the variables_config property to return our mock
        handler._variables_config = mock_config

        # Make write_yaml_file_with_comments raise an exception
        mock_write.side_effect = OSError("Write error")

        # Execute and verify that the exception propagates
        with self.assertRaises(OSError):
            handler.sync_project_variables_config({"var1": "value1"})


class TestResetRecordedVariables(unittest.TestCase):
    @patch("jupyter_deploy.fs_utils.write_yaml_file_with_comments")
    def test_calls_write_updating_required_and_overrides_only(self, mock_write: Mock) -> None:
        # Setup
        project_path = Path("/mock/project")
        manifest = Mock()
        handler = DummyVariablesHandler(project_path=project_path, project_manifest=manifest)

        # Create a mock variables_config with some values
        mock_config = Mock(spec=JupyterDeployVariablesConfig)
        mock_config.required = {"var1": "value1", "var2": "value2"}
        mock_config.required_sensitive = {"var3": "value3"}
        mock_config.overrides = {"var4": "value4"}
        mock_config.defaults = {"var6": "value6"}

        # Patch the variables_config property to return our mock
        handler._variables_config = mock_config

        # Execute
        handler.reset_recorded_variables()

        # Verify
        mock_write.assert_called_once()
        # Get the data that was passed to write_yaml_file_with_comments
        written_data = mock_write.call_args[0][1]

        # Check that all values are None
        self.assertEqual(list(written_data["required"].keys()), list(mock_config.required.keys()))
        for val in written_data["required"].values():
            self.assertIsNone(val)

        self.assertEqual(list(written_data["required_sensitive"].keys()), list(mock_config.required_sensitive.keys()))
        for val in written_data["required_sensitive"].values():
            self.assertIsNone(val)

        # Check that overrides is empty
        self.assertEqual(written_data["overrides"], {})

    @patch("jupyter_deploy.fs_utils.write_yaml_file_with_comments")
    def test_calls_write_with_comments_and_key_order(self, mock_write: Mock) -> None:
        # Setup
        project_path = Path("/mock/project")
        manifest = Mock()
        handler = DummyVariablesHandler(project_path=project_path, project_manifest=manifest)

        # Create a mock variables_config
        mock_config = Mock(spec=JupyterDeployVariablesConfig)
        mock_config.required = {}
        mock_config.required_sensitive = {}
        mock_config.overrides = {}
        mock_config.defaults = {}

        # Patch the variables_config property to return our mock
        handler._variables_config = mock_config

        # Execute
        handler.reset_recorded_variables()

        # Verify that write_yaml_file_with_comments was called with the correct arguments
        mock_write.assert_called_once()
        self.assertEqual(mock_write.call_args[0][0], project_path / constants.VARIABLES_FILENAME)
        self.assertEqual(mock_write.call_args[1]["key_order"], VARIABLES_CONFIG_V1_KEYS_ORDER)
        self.assertEqual(mock_write.call_args[1]["comments"], VARIABLES_CONFIG_V1_COMMENTS)

    @patch("jupyter_deploy.fs_utils.write_yaml_file_with_comments")
    def test_raises_when_write_method_raises(self, mock_write: Mock) -> None:
        # Setup
        project_path = Path("/mock/project")
        manifest = Mock()
        handler = DummyVariablesHandler(project_path=project_path, project_manifest=manifest)

        # Create a mock variables_config
        mock_config = Mock(spec=JupyterDeployVariablesConfig)
        mock_config.required = {}
        mock_config.required_sensitive = {}
        mock_config.overrides = {}
        mock_config.defaults = {}

        # Patch the variables_config property to return our mock
        handler._variables_config = mock_config

        # Make write_yaml_file_with_comments raise an exception
        mock_write.side_effect = OSError("Write error")

        # Execute and verify that the exception propagates
        with self.assertRaises(OSError):
            handler.reset_recorded_variables()


class TestResetRecordedSecrets(unittest.TestCase):
    @patch("jupyter_deploy.fs_utils.write_yaml_file_with_comments")
    def test_calls_write_updating_sensitives_only(self, mock_write: Mock) -> None:
        # Setup
        project_path = Path("/mock/project")
        manifest = Mock()
        handler = DummyVariablesHandler(project_path=project_path, project_manifest=manifest)

        # Create a mock variables_config with some values
        mock_config = Mock(spec=JupyterDeployVariablesConfig)
        mock_config.required = {"var1": "value1", "var2": "value2"}
        mock_config.required_sensitive = {"var3": "value3", "var4": "value4"}
        mock_config.overrides = {"var5": "value55"}
        mock_config.defaults = {"var5": "value5", "var6": "value6"}

        # Create mock for new config with model_dump
        mock_new_config = Mock()
        mock_new_config.model_dump.return_value = {
            "schema_version": 1,
            "required": {"var1": "value1", "var2": "value2"},
            "required_sensitive": {"var3": None, "var4": None},
            "overrides": {"var5": "value55"},
            "defaults": {"var5": "value5", "var6": "value6"},
        }

        # Patch the variables_config property to return our mock
        handler._variables_config = mock_config

        # Execute with patched JupyterDeployVariablesConfigV1
        handler.reset_recorded_secrets()

        # Verify
        mock_write.assert_called_once()
        # Get the data that was passed to write_yaml_file_with_comments
        written_data = mock_write.call_args[0][1]

        # Check the model_dump output was passed to write_yaml_file_with_comments
        self.assertEqual(written_data, mock_new_config.model_dump.return_value)

        # Check that required_sensitive keys are preserved but values are None
        self.assertEqual(list(written_data["required_sensitive"].keys()), list(mock_config.required_sensitive.keys()))
        for val in written_data["required_sensitive"].values():
            self.assertIsNone(val)

    @patch("jupyter_deploy.fs_utils.write_yaml_file_with_comments")
    def test_calls_write_with_comments_and_key_order(self, mock_write: Mock) -> None:
        # Setup
        project_path = Path("/mock/project")
        manifest = Mock()
        handler = DummyVariablesHandler(project_path=project_path, project_manifest=manifest)

        # Create a mock variables_config
        mock_config = Mock(spec=JupyterDeployVariablesConfig)
        mock_config.required = {}
        mock_config.required_sensitive = {}
        mock_config.overrides = {}
        mock_config.defaults = {}

        # Patch the variables_config property to return our mock
        handler._variables_config = mock_config

        # Execute
        handler.reset_recorded_secrets()

        # Verify that write_yaml_file_with_comments was called with the correct arguments
        mock_write.assert_called_once()
        self.assertEqual(mock_write.call_args[0][0], project_path / constants.VARIABLES_FILENAME)
        self.assertEqual(mock_write.call_args[1]["key_order"], VARIABLES_CONFIG_V1_KEYS_ORDER)
        self.assertEqual(mock_write.call_args[1]["comments"], VARIABLES_CONFIG_V1_COMMENTS)

    @patch("jupyter_deploy.fs_utils.write_yaml_file_with_comments")
    def test_raises_when_write_method_raises(self, mock_write: Mock) -> None:
        # Setup
        project_path = Path("/mock/project")
        manifest = Mock()
        handler = DummyVariablesHandler(project_path=project_path, project_manifest=manifest)

        # Create a mock variables_config
        mock_config = Mock(spec=JupyterDeployVariablesConfig)
        mock_config.required = {}
        mock_config.required_sensitive = {}
        mock_config.overrides = {}
        mock_config.defaults = {}

        # Patch the variables_config property to return our mock
        handler._variables_config = mock_config

        # Make write_yaml_file_with_comments raise an exception
        mock_write.side_effect = OSError("Write error")

        # Execute and verify that the exception propagates
        with self.assertRaises(OSError):
            handler.reset_recorded_secrets()



def to_cli_option_name(s: str) -> str:
    """Return name to kebab-case CLI option format.

    Examples:
        FullTitleCase -> full-title-case
        camelCaseName -> camel-case-name
        python_var_name -> python-var-name
        SomeMixed-Case -> some-mixed-case
    """
    if not s:
        return s

    result = []
    prev_char = s[0]
    result.append(prev_char.lower())

    for curr_char in s[1:]:
        # Handle camelCase and TitleCase
        if curr_char.isupper() and prev_char.islower():
            result.append("-")
            result.append(curr_char.lower())
        # Handle underscores
        elif curr_char == "_":
            if prev_char != "-" and prev_char != "_":
                result.append("-")
        # Handle existing hyphens
        elif curr_char == "-":
            if prev_char != "-":
                result.append("-")
        else:
            result.append(curr_char.lower())

        prev_char = curr_char

    return "".join(result).strip("-")


def get_trimmed_header(full_text: str, max_length: int = 120) -> str:
    """Return the full line of text, up to the char limit."""
    if not full_text or max_length <= 0:
        return ""

    no_leading_white_spaces_text = full_text.lstrip()
    trimmed_text = no_leading_white_spaces_text[:max_length]

    split_trimmed_text = trimmed_text.split("\n")
    return split_trimmed_text[0]


def to_list_str(concatenated_list: str, sep: str = ",") -> list[str]:
    """Split the string by the separator, return result."""
    if not concatenated_list:
        return []

    items = concatenated_list.split(sep)
    return items



import unittest
from unittest.mock import Mock, patch

from jupyter_deploy.engine.outdefs import StrTemplateOutputDefinition, TemplateOutputDefinition
from jupyter_deploy.manifest import JupyterDeployCommandV1
from jupyter_deploy.provider.instruction_runner import InterruptInstructionError
from jupyter_deploy.provider.manifest_command_runner import ManifestCommandRunner
from jupyter_deploy.provider.resolved_clidefs import ResolvedCliParameter, StrResolvedCliParameter
from jupyter_deploy.provider.resolved_resultdefs import (
    ListStrResolvedInstructionResult,
    ResolvedInstructionResult,
    StrResolvedInstructionResult,
)


class TestManifestCommandRunner(unittest.TestCase):
    """Tests for ManifestCommandRunner."""

    def get_cmd_def(self) -> JupyterDeployCommandV1:
        return JupyterDeployCommandV1(
            **{
                "cmd": "release-rsus-and-celebrate",
                "sequence": [
                    {
                        "api-name": "hr.questions.ask-nicely",
                        "arguments": [
                            {"api-attribute": "hr-contact", "source": "output", "source-key": "hr_email"},
                            {"api-attribute": "hr-message", "source": "output", "source-key": "message_to_hr"},
                            {"api-attribute": "number-of-rsus", "source": "cli", "source-key": "rsu_count"},
                        ],
                    },
                    {
                        "api-name": "life.celebration.create-event",
                        "arguments": [
                            {"api-attribute": "venue", "source": "output", "source-key": "celebration_venue"},
                            {"api-attribute": "friends", "source": "output", "source-key": "celebration_friends"},
                            {"api-attribute": "budget", "source": "result", "source-key": "[0].rsu_value"},
                        ],
                    },
                ],
                "results": [
                    {"result-name": "hours-of-hangover", "source": "result", "source-key": "[1].side_effect"},
                    {"result-name": "remaining-rsus", "source": "result", "source-key": "[0].rsus_left"},
                    {"result-name": "party", "source": "result", "source-key": "[1].full_party"},
                    {
                        "result-name": "karaoke-winners",
                        "source": "result",
                        "source-key": "[1].winners",
                        "transform": "comma-separated-str-to-list-str",
                    },
                ],
                "updates": [
                    {"variable-name": "available_rsus", "source": "result", "source-key": "[0].rsus_left"},
                    {"variable-name": "party_gang", "source": "result", "source-key": "[1].full_party"},
                    {
                        "variable-name": "karaoke_title_holders",
                        "source": "result",
                        "source-key": "[1].winners",
                        "transform": "comma-separated-str-to-list-str",
                    },
                ],
            }  # type: ignore
        )

    def get_project_outputs(self) -> dict[str, TemplateOutputDefinition]:
        return {
            "hr_email": StrTemplateOutputDefinition(output_name="hr_email", value="nice-hr@company.com"),
            "message_to_hr": StrTemplateOutputDefinition(
                output_name="message_to_hr", value="can I get my RSU vested please?"
            ),
            "celebration_venue": StrTemplateOutputDefinition(
                output_name="celebration_venue", value="incredible-karaoke"
            ),
            "celebration_friends": StrTemplateOutputDefinition(
                output_name="celebration_friends", value="Ross,Rachel,Monika,Phoebe,Joey,Chandler"
            ),
        }

    def get_cli_inputs(self) -> dict[str, ResolvedCliParameter]:
        return {
            "rsu_count": StrResolvedCliParameter(
                parameter_name="number-of-rsus", value="100"
            )  # use int class type when we introduce it
        }

    def get_mocked_first_instruction_results(self) -> dict[str, ResolvedInstructionResult]:
        return {
            "rsus_left": StrResolvedInstructionResult(result_name="rsus_left", value="300"),
            "rsu_value": StrResolvedInstructionResult(result_name="rsu_value", value="2000"),
        }

    def get_mocked_second_instruction_results(self) -> dict[str, ResolvedInstructionResult]:
        return {
            "side_effect": StrResolvedInstructionResult(result_name="side_effect", value="24"),
            "full_party": ListStrResolvedInstructionResult(
                result_name="full_party", value=["Ross", "Rachel", "Monika", "Phoebe", "Joey", "Chandler", "You"]
            ),
            "winners": StrResolvedInstructionResult(result_name="winners", value="Ross,Rachel"),
        }

    def run_command_sequence_setup(self) -> tuple[JupyterDeployCommandV1, Mock, Mock, Mock]:
        """Setup common test fixtures for command sequence execution."""
        cmd = self.get_cmd_def()
        mock_output_defs = self.get_project_outputs()
        mock_cliparam_defs = self.get_cli_inputs()

        console_mock = Mock()
        output_handler_mock = Mock()
        variable_handler_mock = Mock()
        output_handler_mock.get_full_project_outputs.return_value = mock_output_defs

        mock_result_1 = self.get_mocked_first_instruction_results()
        mock_result_2 = self.get_mocked_second_instruction_results()

        mock_runner = Mock()
        mock_runner.execute_instruction.side_effect = [mock_result_1, mock_result_2]

        with patch(
            "jupyter_deploy.provider.instruction_runner_factory.InstructionRunnerFactory.get_provider_instruction_runner",
            return_value=mock_runner,
        ):
            runner = ManifestCommandRunner(
                console=console_mock, output_handler=output_handler_mock, variable_handler=variable_handler_mock
            )
            success, results = runner.run_command_sequence(cmd, mock_cliparam_defs)

        return cmd, console_mock, output_handler_mock, variable_handler_mock

    def test_init_should_not_instantiate_any_provider_runner(self) -> None:
        # Arrange
        console_mock = Mock()
        output_handler_mock = Mock()

        # Act
        runner = ManifestCommandRunner(
            console=console_mock, output_handler=output_handler_mock, variable_handler=Mock()
        )

        # Assert
        self.assertEqual(runner._console, console_mock)
        self.assertEqual(runner._output_handler, output_handler_mock)

    @patch(
        "jupyter_deploy.provider.instruction_runner_factory.InstructionRunnerFactory.get_provider_instruction_runner"
    )
    def test_run_cmd_sequence_should_run_instructions_and_return_result(
        self, mock_get_provider_instruction_runner: Mock
    ) -> None:
        # Arrange
        cmd = self.get_cmd_def()
        mock_output_defs = self.get_project_outputs()
        mock_cliparam_defs = self.get_cli_inputs()

        console_mock = Mock()
        output_handler_mock = Mock()
        output_handler_mock.get_full_project_outputs.return_value = mock_output_defs

        mock_runner = Mock()
        mock_get_provider_instruction_runner.return_value = mock_runner

        mock_result_1 = self.get_mocked_first_instruction_results()
        mock_result_2 = self.get_mocked_second_instruction_results()

        # Set up the mock to return different results for different calls
        mock_runner.execute_instruction.side_effect = [mock_result_1, mock_result_2]

        # Act
        runner = ManifestCommandRunner(
            console=console_mock, output_handler=output_handler_mock, variable_handler=Mock()
        )
        success, results = runner.run_command_sequence(cmd, mock_cliparam_defs)

        # Assert
        self.assertTrue(success)
        self.assertEqual(mock_get_provider_instruction_runner.call_count, 2)
        self.assertEqual(mock_runner.execute_instruction.call_count, 2)

        # Check that the results were properly indexed and returned
        self.assertEqual(results["[0].rsus_left"].value, "300")
        self.assertEqual(results["[0].rsu_value"].value, "2000")
        self.assertEqual(results["[1].side_effect"].value, "24")
        self.assertEqual(
            results["[1].full_party"].value, ["Ross", "Rachel", "Monika", "Phoebe", "Joey", "Chandler", "You"]
        )

    @patch(
        "jupyter_deploy.provider.instruction_runner_factory.InstructionRunnerFactory.get_provider_instruction_runner"
    )
    def test_run_cmd_sequence_should_resolve_all_types_of_args(
        self, mock_get_provider_instruction_runner: Mock
    ) -> None:
        # Arrange
        cmd = self.get_cmd_def()
        mock_output_defs = self.get_project_outputs()
        mock_cliparam_defs = self.get_cli_inputs()

        console_mock = Mock()
        output_handler_mock = Mock()
        output_handler_mock.get_full_project_outputs.return_value = mock_output_defs

        mock_result_1 = self.get_mocked_first_instruction_results()
        mock_result_2 = self.get_mocked_second_instruction_results()
        mock_runner = Mock()
        mock_get_provider_instruction_runner.return_value = mock_runner
        mock_runner.execute_instruction.side_effect = [mock_result_1, mock_result_2]

        # Act
        runner = ManifestCommandRunner(
            console=console_mock, output_handler=output_handler_mock, variable_handler=Mock()
        )
        success, results = runner.run_command_sequence(cmd, mock_cliparam_defs)

        # Assert success
        self.assertTrue(success)

        # Assert
        output_handler_mock.get_full_project_outputs.assert_called()
        self.assertEqual(output_handler_mock.get_full_project_outputs.call_count, 2)

        # Check that the resolved arguments were passed correctly to the execute_instruction method
        calls = mock_runner.execute_instruction.call_args_list

        # First instruction call
        first_call_args = calls[0][1]
        self.assertEqual(first_call_args["instruction_name"], "hr.questions.ask-nicely")
        self.assertIn("hr-contact", first_call_args["resolved_arguments"])
        self.assertIn("hr-message", first_call_args["resolved_arguments"])
        self.assertIn("number-of-rsus", first_call_args["resolved_arguments"])
        self.assertEqual(first_call_args["resolved_arguments"]["hr-contact"].value, "nice-hr@company.com")
        self.assertEqual(first_call_args["resolved_arguments"]["hr-message"].value, "can I get my RSU vested please?")
        self.assertEqual(first_call_args["resolved_arguments"]["number-of-rsus"].value, "100")

        # Second instruction call
        second_call_args = calls[1][1]
        self.assertEqual(second_call_args["instruction_name"], "life.celebration.create-event")
        self.assertIn("venue", second_call_args["resolved_arguments"])
        self.assertIn("friends", second_call_args["resolved_arguments"])
        self.assertIn("budget", second_call_args["resolved_arguments"])
        self.assertEqual(second_call_args["resolved_arguments"]["venue"].value, "incredible-karaoke")
        self.assertEqual(
            second_call_args["resolved_arguments"]["friends"].value, "Ross,Rachel,Monika,Phoebe,Joey,Chandler"
        )
        self.assertEqual(second_call_args["resolved_arguments"]["budget"].value, "2000")

    @patch(
        "jupyter_deploy.provider.instruction_runner_factory.InstructionRunnerFactory.get_provider_instruction_runner"
    )
    def test_run_cmd_sequence_should_call_factory_runner(self, mock_get_provider_instruction_runner: Mock) -> None:
        # Arrange
        cmd = self.get_cmd_def()
        mock_output_defs = self.get_project_outputs()
        mock_cliparam_defs = self.get_cli_inputs()

        console_mock = Mock()
        output_handler_mock = Mock()
        output_handler_mock.get_full_project_outputs.return_value = mock_output_defs

        mock_result_1 = self.get_mocked_first_instruction_results()
        mock_result_2 = self.get_mocked_second_instruction_results()
        mock_runner = Mock()
        mock_get_provider_instruction_runner.return_value = mock_runner
        mock_runner.execute_instruction.side_effect = [mock_result_1, mock_result_2]

        # Act
        runner = ManifestCommandRunner(
            console=console_mock, output_handler=output_handler_mock, variable_handler=Mock()
        )
        success, results = runner.run_command_sequence(cmd, mock_cliparam_defs)

        # Assert
        self.assertTrue(success)

        # Assert
        # Verify that the factory was called with the correct API names
        mock_get_provider_instruction_runner.assert_any_call("hr.questions.ask-nicely", output_handler_mock)
        mock_get_provider_instruction_runner.assert_any_call("life.celebration.create-event", output_handler_mock)
        self.assertEqual(mock_get_provider_instruction_runner.call_count, 2)

    @patch(
        "jupyter_deploy.provider.instruction_runner_factory.InstructionRunnerFactory.get_provider_instruction_runner"
    )
    def test_get_result_value_returns_transformed_result(self, mock_get_provider_instruction_runner: Mock) -> None:
        # Arrange
        cmd = self.get_cmd_def()
        mock_output_defs = self.get_project_outputs()
        mock_cliparam_defs = self.get_cli_inputs()

        console_mock = Mock()
        output_handler_mock = Mock()
        output_handler_mock.get_full_project_outputs.return_value = mock_output_defs

        mock_result_1 = self.get_mocked_first_instruction_results()
        mock_result_2 = self.get_mocked_second_instruction_results()
        mock_runner = Mock()
        mock_get_provider_instruction_runner.return_value = mock_runner
        mock_runner.execute_instruction.side_effect = [mock_result_1, mock_result_2]

        # Act
        runner = ManifestCommandRunner(
            console=console_mock, output_handler=output_handler_mock, variable_handler=Mock()
        )
        success, results = runner.run_command_sequence(cmd, mock_cliparam_defs)

        # Verify results are available
        self.assertTrue(success)
        self.assertEqual(results["[0].rsus_left"].value, "300")

        # Test get_result_value with string result
        hours_of_hangover = runner.get_result_value(cmd, "hours-of-hangover", str)
        self.assertEqual(hours_of_hangover, "24")

        # Test get_result_value with list str result
        party = runner.get_result_value(cmd, "party", list[str])
        self.assertEqual(party, ["Ross", "Rachel", "Monika", "Phoebe", "Joey", "Chandler", "You"])
        winners = runner.get_result_value(cmd, "karaoke-winners", list[str])
        self.assertEqual(winners, ["Ross", "Rachel"])

        # Test with different expected type
        with self.assertRaises(TypeError):
            runner.get_result_value(cmd, "hours-of-hangover", list)
        with self.assertRaises(TypeError):
            runner.get_result_value(cmd, "party", dict)
        with self.assertRaises(TypeError):
            runner.get_result_value(cmd, "karaoke-winners", int)

    @patch(
        "jupyter_deploy.provider.instruction_runner_factory.InstructionRunnerFactory.get_provider_instruction_runner"
    )
    def test_get_result_value_raises_error_for_invalid_result(self, mock_get_provider_instruction_runner: Mock) -> None:
        # Arrange
        cmd = self.get_cmd_def()
        mock_output_defs = self.get_project_outputs()
        mock_cliparam_defs = self.get_cli_inputs()

        console_mock = Mock()
        output_handler_mock = Mock()
        output_handler_mock.get_full_project_outputs.return_value = mock_output_defs

        mock_result_1 = self.get_mocked_first_instruction_results()
        mock_result_2 = self.get_mocked_second_instruction_results()
        mock_runner = Mock()
        mock_get_provider_instruction_runner.return_value = mock_runner
        mock_runner.execute_instruction.side_effect = [mock_result_1, mock_result_2]

        # Act
        runner = ManifestCommandRunner(
            console=console_mock, output_handler=output_handler_mock, variable_handler=Mock()
        )
        success, results = runner.run_command_sequence(cmd, mock_cliparam_defs)

        # Assert
        self.assertTrue(success)

        # Test with non-existent result name
        with self.assertRaises(StopIteration):
            runner.get_result_value(cmd, "non-existent-result", str)

        # Create a modified command with invalid source key
        modified_cmd = self.get_cmd_def()
        invalid_result = JupyterDeployCommandV1.model_validate(
            {
                "cmd": "test",
                "sequence": [],
                "results": [{"result-name": "invalid-result", "source": "result", "source-key": "[99].invalid"}],
            }
        ).results[0]  # type: ignore

        if modified_cmd.results is None:
            modified_cmd.results = [invalid_result]
        else:
            modified_cmd.results.append(invalid_result)

        # Test with invalid source key
        with self.assertRaises(KeyError):
            runner.get_result_value(modified_cmd, "invalid-result", str)

    @patch(
        "jupyter_deploy.provider.instruction_runner_factory.InstructionRunnerFactory.get_provider_instruction_runner"
    )
    def test_update_variables_correctly_sets_values(self, mock_get_provider_instruction_runner: Mock) -> None:
        # Arrange
        cmd = self.get_cmd_def()
        mock_output_defs = self.get_project_outputs()
        mock_cliparam_defs = self.get_cli_inputs()

        console_mock = Mock()
        output_handler_mock = Mock()
        variable_handler_mock = Mock()
        output_handler_mock.get_full_project_outputs.return_value = mock_output_defs

        mock_result_1 = self.get_mocked_first_instruction_results()
        mock_result_2 = self.get_mocked_second_instruction_results()
        mock_runner = Mock()
        mock_get_provider_instruction_runner.return_value = mock_runner
        mock_runner.execute_instruction.side_effect = [mock_result_1, mock_result_2]

        # Act
        runner = ManifestCommandRunner(
            console=console_mock, output_handler=output_handler_mock, variable_handler=variable_handler_mock
        )
        success, results = runner.run_command_sequence(cmd, mock_cliparam_defs)
        self.assertTrue(success)
        runner.update_variables(cmd)

        # Assert
        # Verify that variable_handler.update_variable_records was called with expected values
        expected_values = {
            "available_rsus": "300",
            "party_gang": ["Ross", "Rachel", "Monika", "Phoebe", "Joey", "Chandler", "You"],
            "karaoke_title_holders": ["Ross", "Rachel"],
        }
        variable_handler_mock.update_variable_records.assert_called_once_with(expected_values)
        variable_handler_mock.sync_project_variables_config.assert_called_once_with(expected_values)

    @patch(
        "jupyter_deploy.provider.instruction_runner_factory.InstructionRunnerFactory.get_provider_instruction_runner"
    )
    def test_update_variables_with_no_updates(self, mock_get_provider_instruction_runner: Mock) -> None:
        # Arrange
        cmd = self.get_cmd_def()
        cmd.updates = None  # Remove updates from the command

        mock_output_defs = self.get_project_outputs()
        mock_cliparam_defs = self.get_cli_inputs()

        console_mock = Mock()
        output_handler_mock = Mock()
        variable_handler_mock = Mock()
        output_handler_mock.get_full_project_outputs.return_value = mock_output_defs

        mock_result_1 = self.get_mocked_first_instruction_results()
        mock_result_2 = self.get_mocked_second_instruction_results()
        mock_runner = Mock()
        mock_get_provider_instruction_runner.return_value = mock_runner
        mock_runner.execute_instruction.side_effect = [mock_result_1, mock_result_2]

        # Act
        runner = ManifestCommandRunner(
            console=console_mock, output_handler=output_handler_mock, variable_handler=variable_handler_mock
        )
        success, results = runner.run_command_sequence(cmd, mock_cliparam_defs)
        self.assertTrue(success)
        runner.update_variables(cmd)

        # Assert - variable_handler.update_variable_records should not be called
        variable_handler_mock.update_variable_records.assert_not_called()
        variable_handler_mock.sync_project_variables_config.assert_not_called()

    @patch(
        "jupyter_deploy.provider.instruction_runner_factory.InstructionRunnerFactory.get_provider_instruction_runner"
    )
    def test_update_variables_raises_error_for_invalid_source_key(
        self, mock_get_provider_instruction_runner: Mock
    ) -> None:
        """Test that update_variables raises an error for invalid source keys."""
        # Arrange
        cmd = self.get_cmd_def()
        # Modify the command to include an update with an invalid source key
        invalid_update = JupyterDeployCommandV1.model_validate(
            {
                "cmd": "test",
                "sequence": [],
                "updates": [{"variable-name": "invalid_var", "source": "result", "source-key": "[99].invalid"}],
            }
        ).updates[0]  # type: ignore

        if cmd.updates is None:
            cmd.updates = [invalid_update]
        else:
            cmd.updates.append(invalid_update)

        mock_output_defs = self.get_project_outputs()
        mock_cliparam_defs = self.get_cli_inputs()

        console_mock = Mock()
        output_handler_mock = Mock()
        variable_handler_mock = Mock()
        output_handler_mock.get_full_project_outputs.return_value = mock_output_defs

        mock_result_1 = self.get_mocked_first_instruction_results()
        mock_result_2 = self.get_mocked_second_instruction_results()
        mock_runner = Mock()
        mock_get_provider_instruction_runner.return_value = mock_runner
        mock_runner.execute_instruction.side_effect = [mock_result_1, mock_result_2]

        # Act
        runner = ManifestCommandRunner(
            console=console_mock, output_handler=output_handler_mock, variable_handler=variable_handler_mock
        )
        success, results = runner.run_command_sequence(cmd, mock_cliparam_defs)

        # Assert
        self.assertTrue(success)

        # Assert - should raise KeyError for invalid source key
        with self.assertRaises(KeyError):
            runner.update_variables(cmd)

    @patch(
        "jupyter_deploy.provider.instruction_runner_factory.InstructionRunnerFactory.get_provider_instruction_runner"
    )
    @patch("typer.Abort")
    def test_interrupt_instruction_error_calls_typer_abort(
        self, mock_typer_abort: Mock, mock_get_provider_instruction_runner: Mock
    ) -> None:
        # Arrange
        cmd = self.get_cmd_def()
        mock_output_defs = self.get_project_outputs()
        mock_cliparam_defs = self.get_cli_inputs()

        console_mock = Mock()
        output_handler_mock = Mock()
        output_handler_mock.get_full_project_outputs.return_value = mock_output_defs

        # Configure the mock runner to raise InterruptInstructionError
        mock_runner = Mock()
        mock_runner.execute_instruction.side_effect = InterruptInstructionError()
        mock_get_provider_instruction_runner.return_value = mock_runner

        # Act
        runner = ManifestCommandRunner(
            console=console_mock, output_handler=output_handler_mock, variable_handler=Mock()
        )
        success, results = runner.run_command_sequence(cmd, mock_cliparam_defs)

        # Assert
        self.assertFalse(success)
        mock_typer_abort.assert_called_once()



import json
from abc import ABC, abstractmethod
from typing import Generic, TypeVar  # noqa: UP035

from pydantic import BaseModel, ConfigDict

from jupyter_deploy import type_utils
from jupyter_deploy.engine.outdefs import (
    ListStrTemplateOutputDefinition,
    StrTemplateOutputDefinition,
    TemplateOutputDefinition,
)
from jupyter_deploy.engine.terraform.tf_types import TerraformType

V = TypeVar("V")


class ParsedOutputsDotTf(BaseModel):
    model_config = ConfigDict(extra="allow")
    output: list[dict[str, dict]]


class TerraformOutputDefinition(BaseModel, Generic[V], ABC):
    """Terraform-specific wrapper class for variable definition."""

    model_config = ConfigDict(extra="allow")
    output_name: str
    tf_type: TerraformType
    description: str = ""
    value: V | None = None
    sensitive: bool = False

    @abstractmethod
    def to_template_definition(self) -> TemplateOutputDefinition:
        """Return the equivalent instance of the TemplateOutputDefinition subclass."""
        pass


class TerraformStrOutputDefinition(TerraformOutputDefinition[str]):
    """Terraform wrapper class for variable definition of type 'string'."""

    tf_type: TerraformType = TerraformType.STRING

    def to_template_definition(self) -> TemplateOutputDefinition:
        return StrTemplateOutputDefinition(**self.model_dump())


class TerraformListStrOutputDefinition(TerraformOutputDefinition[list[str]]):
    """Terraform wrapper class for variable definition of type 'list(string)'."""

    tf_type: TerraformType = TerraformType.LIST_STR

    def to_template_definition(self) -> TemplateOutputDefinition:
        return ListStrTemplateOutputDefinition(**self.model_dump())


def create_tf_output_definition(parsed_config: dict) -> TerraformOutputDefinition:
    """Return an instance of the corresponding TerraformOutput class."""
    tf_type = parsed_config.get("type")
    if tf_type == TerraformType.STRING:
        return TerraformStrOutputDefinition(**parsed_config)
    elif tf_type == TerraformType.LIST_STR or type_utils.is_list_str_repr(tf_type):
        return TerraformListStrOutputDefinition(**parsed_config)

    raise NotImplementedError(f"No terraform class found for type: {tf_type}.")


def parse_output_cmd_result(content: str) -> dict[str, TerraformOutputDefinition]:
    """Parse the result of the output cmd, return a dict of output_name->config."""
    output_defs: dict[str, TerraformOutputDefinition] = {}

    output_dict = json.loads(content)
    if not isinstance(output_dict, dict):
        raise RuntimeError("Expected a dict from terraform output --json command")

    for output_name, output_config in output_dict.items():
        if not isinstance(output_name, str):
            raise RuntimeError(f"Expected output name to be a str, got: {type(output_name)}")
        if not isinstance(output_config, dict):
            raise RuntimeError(f"Expected output config to be a dict, got {type(output_config)}")

        full_config = {**output_config, "output_name": output_name}
        out_def = create_tf_output_definition(full_config)
        output_defs.update({output_name: out_def})

    return output_defs



from abc import ABC, abstractmethod
from pathlib import Path

from jupyter_deploy.engine.engine_variables import EngineVariablesHandler
from jupyter_deploy.engine.enum import EngineType
from jupyter_deploy.engine.vardefs import TemplateVariableDefinition
from jupyter_deploy.manifest import JupyterDeployManifest


class EngineConfigHandler(ABC):
    def __init__(
        self,
        project_path: Path,
        project_manifest: JupyterDeployManifest,
        engine: EngineType,
        variables_handler: EngineVariablesHandler,
    ) -> None:
        """Instantiate the base handler for `jd config` command."""
        self.project_path = project_path
        self.project_manifest = project_manifest
        self.engine = engine
        self.variables_handler = variables_handler

    @abstractmethod
    def has_recorded_variables(self) -> bool:
        """Return True if the file generated by jd config to record variable values exists in the dir."""
        pass

    @abstractmethod
    def verify_requirements(self) -> bool:
        """Returns True if all required dependencies for this template are installed."""
        pass

    @abstractmethod
    def verify_preset_exists(self, preset_name: str) -> bool:
        """Return True if the requested preset is defined by this template."""
        pass

    @abstractmethod
    def list_presets(self) -> list[str]:
        """Return the list of default presets that this template supports."""
        pass

    @abstractmethod
    def configure(
        self, preset_name: str | None = None, variable_overrides: dict[str, TemplateVariableDefinition] | None = None
    ) -> bool:
        """Execute commands to set the values of the variables, return True if succeeded."""
        pass

    @abstractmethod
    def reset_recorded_variables(self) -> None:
        """Delete the file in the project dir where the previous inputs were recorded."""
        pass

    @abstractmethod
    def reset_recorded_secrets(self) -> None:
        """Delete the file in the project dir where the previous secrets were recorded."""
        pass

    @abstractmethod
    def record(self, record_vars: bool = False, record_secrets: bool = False) -> None:
        """Save the values of the variables and/or secrets to disk in the project dir."""
        pass



from rich.table import Table

from jupyter_deploy.engine.engine_outputs import EngineOutputsHandler
from jupyter_deploy.engine.engine_variables import EngineVariablesHandler
from jupyter_deploy.engine.enum import EngineType
from jupyter_deploy.engine.terraform import tf_outputs, tf_variables
from jupyter_deploy.handlers.base_project_handler import BaseProjectHandler


class ShowHandler(BaseProjectHandler):
    """Handler for displaying project information and outputs."""

    _outputs_handler: EngineOutputsHandler
    _variables_handler: EngineVariablesHandler

    def __init__(self) -> None:
        """Initialize the show handler."""
        super().__init__()

        if self.engine == EngineType.TERRAFORM:
            self._outputs_handler = tf_outputs.TerraformOutputsHandler(
                project_path=self.project_path,
                project_manifest=self.project_manifest,
            )
            self._variables_handler = tf_variables.TerraformVariablesHandler(
                project_path=self.project_path,
                project_manifest=self.project_manifest,
            )
        else:
            raise NotImplementedError(f"ShowHandler implementation not found for engine: {self.engine}")

    def show_project_info(self, show_info: bool = True, show_outputs: bool = True, show_variables: bool = True) -> None:
        """Display comprehensive project information."""
        console = self.get_console()

        if show_info:
            console.line()
            self._show_project_basic_info()
        if show_variables:
            console.line()
            self._show_project_variables()
        if show_outputs:
            console.line()
            self._show_project_outputs()

    def _show_project_basic_info(self) -> None:
        """Display basic project information."""
        console = self.get_console()

        console.print("Jupyter Deploy Project Information", style="bold cyan")
        console.line()

        table = Table(show_header=True, header_style="bold magenta")
        table.add_column("Property", style="cyan", no_wrap=True)
        table.add_column("Value", style="white")

        table.add_row("Project Path", str(self.project_path))
        table.add_row("Engine", self.engine.value)
        table.add_row("Template Name", self.project_manifest.template.name)
        table.add_row("Template Version", self.project_manifest.template.version)

        console.print(table)
        console.line()

    def _show_project_outputs(self) -> None:
        """Display project outputs if they exist."""
        console = self.get_console()
        try:
            outputs = self._outputs_handler.get_full_project_outputs()
        except Exception as e:
            console.print(f":x: Could not retrieve outputs: {str(e)}", style="red")
            console.line()
            return

        if not outputs:
            console.print(":warning: No outputs available.", style="yellow")
            console.print("This is normal if the project has not been deployed yet.", style="yellow")
            console.line()
            return

        console.print("Project Outputs", style="bold cyan")
        console.line()

        output_table = Table(show_header=True, header_style="bold magenta")
        output_table.add_column("Output Name", style="cyan", no_wrap=True)
        output_table.add_column("Value", style="white")
        output_table.add_column("Description", style="dim")

        for output_name, output_def in outputs.items():
            description = getattr(output_def, "description", "") or "No description"
            value = str(output_def.value) if hasattr(output_def, "value") and output_def.value is not None else "N/A"
            output_table.add_row(output_name, value, description)

        console.print(output_table)

    def _show_project_variables(self) -> None:
        """Display project variables based on the values set in 'variables.yaml'."""
        console = self.get_console()
        try:
            self._variables_handler.sync_engine_varfiles_with_project_variables_config()
            variables = self._variables_handler.get_template_variables()
        except Exception as e:
            console.print(f":x: Could not retrieve variables: {str(e)}", style="red")
            return

        if not variables:
            console.print(":x: No variables available.", style="red")
            return

        console.print("Project Variables", style="bold cyan")
        console.line()

        variables_table = Table(show_header=True, header_style="bold magenta")
        variables_table.add_column("Variable Name", style="cyan", no_wrap=True)
        variables_table.add_column("Assigned Value", style="white")
        variables_table.add_column("Description", style="dim")

        for variable_name, variable_def in variables.items():
            description = variable_def.get_cli_description()
            sensitive = variable_def.sensitive
            if not sensitive:
                assigned_value = str(variable_def.assigned_value) if hasattr(variable_def, "assigned_value") else None
            else:
                assigned_value = "****"
            variables_table.add_row(variable_name, assigned_value, description)

        console.print(variables_table)



import unittest
from unittest.mock import Mock

import botocore.exceptions

from jupyter_deploy.api.aws.ssm.ssm_session import describe_instance_information


class TestDescribeInstanceInformation(unittest.TestCase):
    """Tests for the describe_instance_information function."""

    def setUp(self) -> None:
        """Set up test fixtures."""
        self.mock_ssm_client = Mock()
        self.instance_id = "i-1234567890abcdef0"

    def test_successful_describe_instance_information(self) -> None:
        """Test that describe_instance_information returns the first item when information is available."""
        # Setup
        # Using dict instead of strict typing for mocking purposes
        instance_info = {
            "InstanceId": self.instance_id,
            "PingStatus": "Online",
            # LastPingDateTime should be a datetime object but Mock will handle it
            "AgentVersion": "1.0.0.0",
            "IsLatestVersion": True,
            "PlatformType": "Linux",
            "PlatformName": "Amazon Linux",
            "PlatformVersion": "2",
            "ResourceType": "EC2Instance",
        }
        self.mock_ssm_client.describe_instance_information.return_value = {"InstanceInformationList": [instance_info]}

        # Execute
        result = describe_instance_information(self.mock_ssm_client, self.instance_id)

        # Assert
        self.mock_ssm_client.describe_instance_information.assert_called_once_with(
            Filters=[{"Key": "InstanceIds", "Values": [self.instance_id]}]
        )
        self.assertEqual(result, instance_info)

    def test_empty_instance_information_list(self) -> None:
        """Test that describe_instance_information raises ValueError when no information is available."""
        # Setup
        self.mock_ssm_client.describe_instance_information.return_value = {"InstanceInformationList": []}

        # Execute & Assert
        with self.assertRaises(ValueError) as context:
            describe_instance_information(self.mock_ssm_client, self.instance_id)

        self.assertEqual(str(context.exception), "SSM:DescribeInstanceInformation returned an empty list")
        self.mock_ssm_client.describe_instance_information.assert_called_once_with(
            Filters=[{"Key": "InstanceIds", "Values": [self.instance_id]}]
        )

    def test_api_call_exception_bubbles_up(self) -> None:
        """Test that exceptions from the API call are not swallowed but bubble up."""
        # Setup
        error = botocore.exceptions.ClientError(
            {
                "Error": {
                    "Code": "InvalidInstanceId",
                    "Message": "The instance ID is not valid",
                }
            },
            "DescribeInstanceInformation",
        )
        self.mock_ssm_client.describe_instance_information.side_effect = error

        # Execute & Assert
        with self.assertRaises(botocore.exceptions.ClientError) as context:
            describe_instance_information(self.mock_ssm_client, self.instance_id)

        # Verify the exception is the same one we set up
        self.assertIs(context.exception, error)
        self.mock_ssm_client.describe_instance_information.assert_called_once_with(
            Filters=[{"Key": "InstanceIds", "Values": [self.instance_id]}]
        )



from unittest.mock import patch

import pytest

from jupyter_deploy.engine.enum import EngineType
from jupyter_deploy.handlers.project.open_handler import OpenHandler
from jupyter_deploy.manifest import JupyterDeployManifestV1


@pytest.fixture
def mock_manifest() -> JupyterDeployManifestV1:
    """Create a mock manifest."""
    return JupyterDeployManifestV1(
        **{  # type: ignore
            "schema_version": 1,
            "template": {
                "name": "mock-template-name",
                "engine": "terraform",
                "version": "1.0.0",
            },
            "values": [{"name": "open_url", "source": "output", "source-key": "jupyter_url"}],
        }
    )


class TestOpenHandler:
    def test_init(self, mock_manifest: JupyterDeployManifestV1) -> None:
        """Test that the OpenHandler initializes correctly."""
        with patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest") as mock_retrieve_manifest:
            mock_retrieve_manifest.return_value = mock_manifest
            handler = OpenHandler()
            assert handler._handler is not None
            assert handler.engine == EngineType.TERRAFORM
            assert handler.project_manifest == mock_manifest

    def test_open_success(self, mock_manifest: JupyterDeployManifestV1) -> None:
        """Test that open opens the URL in a web browser, and outputs the URL and cookies help message."""
        with patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest") as mock_retrieve_manifest:
            mock_retrieve_manifest.return_value = mock_manifest
            handler = OpenHandler()
            with (
                patch.object(handler._handler, "get_url", return_value="https://example.com/jupyter") as mock_get_url,
                patch("webbrowser.open", return_value=True) as mock_open,
                patch.object(handler.console, "print") as mock_print,
            ):
                handler.open()
                mock_get_url.assert_called_once()
                mock_open.assert_called_once_with("https://example.com/jupyter", new=2)
                assert mock_print.call_count >= 2
                assert "Opening Jupyter" in mock_print.call_args_list[0][0][0]

    def test_open_url_error(self, mock_manifest: JupyterDeployManifestV1) -> None:
        """Test that open_url handles errors when opening the URL."""
        with patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest") as mock_retrieve_manifest:
            mock_retrieve_manifest.return_value = mock_manifest
            handler = OpenHandler()

            with (
                patch.object(handler._handler, "get_url", return_value="https://example.com/jupyter") as mock_get_url,
                patch("webbrowser.open", return_value=False) as mock_open,
                patch.object(handler.console, "print") as mock_print,
            ):
                handler.open()
                mock_get_url.assert_called_once()
                mock_open.assert_called_once_with("https://example.com/jupyter", new=2)
                assert mock_print.call_count == 2
                assert "Failed to open URL" in mock_print.call_args_list[1][0][0]

    def test_open_url_insecure(self, mock_manifest: JupyterDeployManifestV1) -> None:
        """Test that open_url doesn't open non-HTTPS urls."""
        with patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest") as mock_retrieve_manifest:
            mock_retrieve_manifest.return_value = mock_manifest
            handler = OpenHandler()
            with (
                patch.object(handler._handler, "get_url", return_value="http://example.com/jupyter") as mock_get_url,
                patch("webbrowser.open") as mock_open,
                patch.object(handler.console, "print") as mock_print,
            ):
                handler.open()
                mock_get_url.assert_called_once()
                mock_open.assert_not_called()
                mock_print.assert_called_once()
                assert "Insecure URL detected" in mock_print.call_args[0][0]



import unittest
from pathlib import Path
from typing import Any
from unittest.mock import Mock, patch

from jupyter_deploy.engine.terraform import tf_vardefs, tf_varfiles


class TestParseVariablesDotTfContent(unittest.TestCase):
    variables_tf_content: str

    @classmethod
    def setUpClass(cls) -> None:
        mock_variables_path = Path(__file__).parent / "mock_variables.tf"
        with open(mock_variables_path) as f:
            cls.variables_tf_content = f.read()

    def test_can_parse_empty_content(self) -> None:
        result = tf_varfiles.parse_variables_dot_tf_content("")
        self.assertEqual({}, result)

    @patch("hcl2.loads")
    def test_calls_hcl2_loads(self, mock_loads: Mock) -> None:
        mock_loads.return_value = {"variable": []}
        tf_varfiles.parse_variables_dot_tf_content(self.variables_tf_content)
        mock_loads.assert_called_once_with(self.variables_tf_content)

    @patch("hcl2.loads")
    def test_parses_valid_variables_tf(self, mock_loads: Mock) -> None:
        # Create a mock return value that simulates the parsed variables.tf content
        mock_parsed = {
            "variable": [
                {
                    "some_string_value": {
                        "description": "For example the instance type.\n\nRecommended: t3.medium",
                        "type": "string",
                    }
                },
                {
                    "some_int_value": {
                        "description": "For example the size of the disk in GB.\n\nRecommended: 30",
                        "type": "number",
                    }
                },
            ]
        }
        mock_loads.return_value = mock_parsed
        result = tf_varfiles.parse_variables_dot_tf_content(self.variables_tf_content)

        # Verify the result contains the expected variables
        self.assertIn("some_string_value", result)
        self.assertIn("some_int_value", result)

        # Verify the variable definitions have the correct types
        self.assertIsInstance(result["some_string_value"], tf_vardefs.TerraformStrVariableDefinition)
        self.assertIsInstance(result["some_int_value"], tf_vardefs.TerraformNumberVariableDefinition)

        # Verify the descriptions are correctly parsed
        self.assertEqual(
            "For example the instance type.\n\nRecommended: t3.medium", result["some_string_value"].description
        )
        self.assertEqual(
            "For example the size of the disk in GB.\n\nRecommended: 30", result["some_int_value"].description
        )

    def test_parsing_actual_variables_tf_file_works(self) -> None:
        content = self.variables_tf_content
        result = tf_varfiles.parse_variables_dot_tf_content(content)

        # Verify all expected variables are present
        self.assertIn("some_string_value", result)
        self.assertIn("some_int_value", result)
        self.assertIn("some_float_value", result)
        self.assertIn("some_string_value_with_condition", result)
        self.assertIn("some_secret", result)

        # Verify variables have the correct types
        self.assertIsInstance(result["some_string_value"], tf_vardefs.TerraformStrVariableDefinition)
        self.assertIsInstance(result["some_int_value"], tf_vardefs.TerraformNumberVariableDefinition)
        self.assertIsInstance(result["some_float_value"], tf_vardefs.TerraformNumberVariableDefinition)
        self.assertIsInstance(result["some_string_value_with_condition"], tf_vardefs.TerraformStrVariableDefinition)
        self.assertIsInstance(result["some_secret"], tf_vardefs.TerraformStrVariableDefinition)
        self.assertIsInstance(result["some_list_of_string"], tf_vardefs.TerraformListOfStrVariableDefinition)
        self.assertIsInstance(result["some_map_of_sring"], tf_vardefs.TerraformMapOfStrVariableDefinition)

        # Verify descriptions are correctly parsed
        self.assertIn("Recommended: t3.medium", result["some_string_value"].description)
        self.assertIn("For example the size of the disk in GB.", result["some_int_value"].description)
        self.assertIn("Recommended: 0.9", result["some_float_value"].description)

        # Verify sensitive flag is set correctly
        self.assertTrue(result["some_secret"].sensitive)
        self.assertFalse(result["some_string_value"].sensitive)


class TestParseDotTfvarsContentAndAddDefaults(unittest.TestCase):
    variables_tf_content: str
    tfvars_content: str

    @classmethod
    def setUpClass(cls) -> None:
        mock_variables_path = Path(__file__).parent / "mock_variables.tf"
        with open(mock_variables_path) as f:
            cls.variables_tf_content = f.read()

        mock_tfvars_path = Path(__file__).parent / "mock_defaults.tfvars"
        with open(mock_tfvars_path) as f:
            cls.tfvars_content = f.read()

    def test_can_parse_empty_content(self) -> None:
        variable_defs: dict[str, tf_vardefs.TerraformVariableDefinition] = {}
        tf_varfiles.parse_dot_tfvars_content_and_add_defaults("", variable_defs)
        self.assertEqual({}, variable_defs)  # Should remain empty

    @patch("hcl2.loads")
    def test_calls_hcl2_loads(self, mock_loads: Mock) -> None:
        mock_loads.return_value = {}
        variable_defs: dict = {}
        tf_varfiles.parse_dot_tfvars_content_and_add_defaults(self.tfvars_content, variable_defs)
        mock_loads.assert_called_once_with(self.tfvars_content)

    @patch("hcl2.loads")
    def test_parse_valid_tfvars_content_and_add(self, mock_loads: Mock) -> None:
        # Create mock variable definitions
        string_var_def = tf_vardefs.TerraformStrVariableDefinition(
            variable_name="some_string_value", description="A string variable"
        )
        int_var_def = tf_vardefs.TerraformNumberVariableDefinition(
            variable_name="some_int_value", description="An int variable"
        )
        unreferenced_var_def = tf_vardefs.TerraformStrVariableDefinition(
            variable_name="unref_value", description="Another string variable"
        )

        variable_defs: dict[str, tf_vardefs.TerraformVariableDefinition] = {
            "some_string_value": string_var_def,
            "some_int_value": int_var_def,
            "unref_value": unreferenced_var_def,
        }

        # Create mock tfvars content
        mock_tfvars = {"some_string_value": "t3.medium", "some_int_value": 30}

        mock_loads.return_value = mock_tfvars
        tf_varfiles.parse_dot_tfvars_content_and_add_defaults(self.tfvars_content, variable_defs)

        # Verify defaults were added to the variable definitions
        self.assertEqual("t3.medium", variable_defs["some_string_value"].default)
        self.assertEqual(30, variable_defs["some_int_value"].default)
        self.assertIsNone(variable_defs["unref_value"].default)
        self.assertTrue(variable_defs["some_string_value"].has_default)
        self.assertTrue(variable_defs["some_int_value"].has_default)
        self.assertFalse(variable_defs["unref_value"].has_default)

    @patch("hcl2.loads")
    def test_skips_vars_not_found_in_variables_tf(self, mock_loads: Mock) -> None:
        # Create mock variable definitions
        string_var_def = tf_vardefs.TerraformStrVariableDefinition(
            variable_name="some_string_value", description="A string variable"
        )

        variable_defs: dict[str, tf_vardefs.TerraformVariableDefinition] = {"some_string_value": string_var_def}

        # Create mock tfvars content with an unrecognized variable
        mock_tfvars = {"some_string_value": "t3.medium", "some_unrecognized_value": "manually-added"}

        mock_loads.return_value = mock_tfvars
        tf_varfiles.parse_dot_tfvars_content_and_add_defaults(self.tfvars_content, variable_defs)

        # Verify only the recognized variable was updated
        self.assertEqual("t3.medium", variable_defs["some_string_value"].default)
        self.assertEqual(1, len(variable_defs))  # No new variables should be added

    @patch("hcl2.loads")
    def test_skips_sensitive_variables(self, mock_loads: Mock) -> None:
        # Create mock variable definitions including a sensitive one
        string_var_def = tf_vardefs.TerraformStrVariableDefinition(
            variable_name="some_string_value", description="A string variable"
        )
        sensitive_var_def = tf_vardefs.TerraformStrVariableDefinition(
            variable_name="some_secret", description="A sensitive variable", sensitive=True
        )

        variable_defs: dict[str, tf_vardefs.TerraformVariableDefinition] = {
            "some_string_value": string_var_def,
            "some_secret": sensitive_var_def,
        }

        # Create mock tfvars content with a sensitive variable
        mock_tfvars = {"some_string_value": "t3.medium", "some_secret": "i-should-not-be-here"}

        mock_loads.return_value = mock_tfvars
        tf_varfiles.parse_dot_tfvars_content_and_add_defaults(self.tfvars_content, variable_defs)

        # Verify only the non-sensitive variable was updated
        self.assertEqual("t3.medium", variable_defs["some_string_value"].default)
        self.assertIsNone(variable_defs["some_secret"].default)  # Sensitive variable should not be updated

    def test_parsing_actual_variables_tf_and_defaults_tfvars_files_works(self) -> None:
        parsed_variable_defs = tf_varfiles.parse_variables_dot_tf_content(self.variables_tf_content)
        tf_varfiles.parse_dot_tfvars_content_and_add_defaults(self.tfvars_content, parsed_variable_defs)

        # Verify defaults were added to the variable definitions
        self.assertEqual("t3.medium", parsed_variable_defs["some_string_value"].default)
        self.assertEqual(30, parsed_variable_defs["some_int_value"].default)
        self.assertEqual(0.9, parsed_variable_defs["some_float_value"].default)
        self.assertIsNone(parsed_variable_defs["some_string_value_with_condition"].default)
        self.assertEqual(["a", "b"], parsed_variable_defs["some_list_of_string"].default)
        self.assertEqual({}, parsed_variable_defs["some_map_of_sring"].default)

        # Verify sensitive variable does not have default set
        self.assertIn("some_secret", parsed_variable_defs)
        self.assertIsNone(parsed_variable_defs["some_secret"].default)

        # Verify unrecognized variable from tfvars is not added
        self.assertNotIn("some_unrecognized_value", parsed_variable_defs)


class TestParseAndUpdateDotTfvarsContent(unittest.TestCase):
    tfvars_content: str

    @classmethod
    def setUpClass(cls) -> None:
        mock_tfvars_path = Path(__file__).parent / "mock_defaults.tfvars"
        with open(mock_tfvars_path) as f:
            cls.tfvars_content = f.read()

    def test_can_parse_empty_content(self) -> None:
        varvalues: dict[str, Any] = {}
        result = tf_varfiles.parse_and_update_dot_tfvars_content("", varvalues)
        self.assertEqual([], result)

    @patch("hcl2.loads")
    def test_calls_hcl2_loads(self, mock_loads: Mock) -> None:
        mock_loads.return_value = {}
        varvalues: dict[str, Any] = {}
        tf_varfiles.parse_and_update_dot_tfvars_content(self.tfvars_content, varvalues)
        mock_loads.assert_called_once_with(self.tfvars_content)

    def test_parsing_actual_tfvars_works(self) -> None:
        varvalues: dict[str, Any] = {"some_list_of_string": ["b", "c", "d"]}
        result = tf_varfiles.parse_and_update_dot_tfvars_content(self.tfvars_content, varvalues)

        self.assertIn('some_string_value = "t3.medium"\n', result)
        self.assertIn("some_int_value = 30\n", result)
        self.assertIn("some_float_value = 0.9\n", result)
        self.assertIn("some_string_value_with_condition = null\n", result)
        self.assertIn('some_list_of_string = [\n"b",\n"c",\n"d",\n]\n', result)
        self.assertIn('some_secret = "i-should-not-be-here"\n', result)
        self.assertIn("some_map_of_sring = {}\n", result)
        self.assertIn('some_unrecognized_value = "manually-added"\n', result)


class TestParseAndRemoveOverriddenVariablesFromContent(unittest.TestCase):
    tfvars_content: str

    @classmethod
    def setUpClass(cls) -> None:
        mock_tfvars_path = Path(__file__).parent / "mock_defaults.tfvars"
        with open(mock_tfvars_path) as f:
            cls.tfvars_content = f.read()

    def test_can_parse_empty_content(self) -> None:
        varnames_to_remove: list[str] = ["some_var"]
        result = tf_varfiles.parse_and_remove_overridden_variables_from_content("", varnames_to_remove)
        self.assertEqual([], result)

    @patch("hcl2.loads")
    @patch("jupyter_deploy.engine.terraform.tf_plan.format_values_for_dot_tfvars")
    def test_calls_hcl2_loads_remove_match_variables(self, mock_format: Mock, mock_loads: Mock) -> None:
        # Mock return values
        mock_loads.return_value = {"var1": "value1", "var2": "value2", "var3": "value3"}
        mock_format.return_value = ["formatted_line1", "formatted_line2"]

        # Variables to remove
        varnames_to_remove = ["var1", "var3"]

        # Call the function
        result = tf_varfiles.parse_and_remove_overridden_variables_from_content(self.tfvars_content, varnames_to_remove)

        # Verify hcl2.loads was called with the content
        mock_loads.assert_called_once_with(self.tfvars_content)

        # Verify the specified variables were removed from the dict
        mock_format.assert_called_once_with({"var2": "value2"})

        # Verify the result is what was returned by format_values_for_dot_tfvars
        self.assertEqual(["formatted_line1", "formatted_line2"], result)

    def test_parsing_actual_tfvars_and_removing_variables(self) -> None:
        # Variables to remove
        varnames_to_remove = ["some_string_value", "some_int_value", "some_secret"]

        # Call the function with actual tfvars content
        result = tf_varfiles.parse_and_remove_overridden_variables_from_content(self.tfvars_content, varnames_to_remove)

        # Verify the result does not contain the removed variables
        result_content = "\n".join(result)
        self.assertNotIn('some_string_value = "t3.medium"', result_content)
        self.assertNotIn("some_int_value = 30", result_content)
        self.assertNotIn('some_secret = "i-should-not-be-here"', result_content)

        # Verify the result still contains other variables
        self.assertIn("some_float_value", result_content)
        self.assertIn("some_string_value_with_condition", result_content)
        self.assertIn("some_list_of_string", result_content)
        self.assertIn("some_map_of_sring", result_content)
        self.assertIn("some_unrecognized_value", result_content)

    @patch("hcl2.loads")
    @patch("jupyter_deploy.engine.terraform.tf_plan.format_values_for_dot_tfvars")
    def test_ignore_set_variables_not_found_in_tfvars(self, mock_format: Mock, mock_loads: Mock) -> None:
        # Mock return values
        mock_loads.return_value = {"var1": "value1", "var2": "value2"}
        mock_format.return_value = ["formatted_line1", "formatted_line2"]

        # Variables to remove, including some not in the tfvars
        varnames_to_remove = ["var1", "non_existent_var1", "non_existent_var2"]

        # Call the function
        result = tf_varfiles.parse_and_remove_overridden_variables_from_content(self.tfvars_content, varnames_to_remove)

        # Verify the non-existent variables were silently ignored
        mock_format.assert_called_once_with({"var2": "value2"})

        # Verify the result is what was returned by format_values_for_dot_tfvars
        self.assertEqual(["formatted_line1", "formatted_line2"], result)



from abc import ABC, abstractmethod
from pathlib import Path

from jupyter_deploy.engine import outdefs
from jupyter_deploy.engine.enum import EngineType
from jupyter_deploy.enum import ValueSource
from jupyter_deploy.manifest import JupyterDeployManifest


class EngineOutputsHandler(ABC):
    def __init__(self, project_path: Path, engine: EngineType, project_manifest: JupyterDeployManifest) -> None:
        """Instantiate the base handler to access output values."""
        self.project_path = project_path
        self.project_manifest = project_manifest
        self.engine = engine

    @abstractmethod
    def get_full_project_outputs(self) -> dict[str, outdefs.TemplateOutputDefinition]:
        """Return the dict of output_name->output_definition.

        This operation presumably requires file system operations and should
        be cached within each OutputsHandler.
        """
        pass

    @abstractmethod
    def get_output_definition(self, output_name: str) -> outdefs.TemplateOutputDefinition:
        """Return the specified output definition.

        This operation may be expensive, the full dict of outputs should
        be cached within each OutputsHandler
        """
        pass

    def get_declared_output_def(self, value_name: str, value_type: type[outdefs.TOD]) -> outdefs.TOD:
        """Return the output defintion for the value declared in the manifest.values section.

        Raises:
            NotImplementedError if the value_name is not declared in the manifest
            ValueError if the value.source is not a template output
            KeyError if the output name declared in the manifest cannot be found
            TypeError if the output def exists but is not of the valid type
        """
        value_def = self.project_manifest.get_declared_value(value_name)
        value_source_type = value_def.get_source_type()
        value_source_key = value_def.source_key

        if value_source_type != ValueSource.TEMPLATE_OUTPUT:
            raise ValueError(
                f"Manifest declared value is not of type '{ValueSource.TEMPLATE_OUTPUT}': {value_source_type}"
            )

        output_defs = self.get_full_project_outputs()
        output_def = outdefs.require_output_def(
            output_defs=output_defs, output_name=value_source_key, output_type=value_type
        )
        return output_def



from jupyter_deploy.engine.engine_outputs import EngineOutputsHandler
from jupyter_deploy.engine.engine_variables import EngineVariablesHandler
from jupyter_deploy.engine.enum import EngineType
from jupyter_deploy.engine.terraform import tf_outputs, tf_variables
from jupyter_deploy.handlers.base_project_handler import BaseProjectHandler
from jupyter_deploy.provider import manifest_command_runner as cmd_runner
from jupyter_deploy.provider.resolved_clidefs import StrResolvedCliParameter


class UsersHandler(BaseProjectHandler):
    """Handler class to manage user access to a jupyter app."""

    _output_handler: EngineOutputsHandler
    _variable_handler: EngineVariablesHandler

    def __init__(self) -> None:
        """Instantiate the Users handler."""
        super().__init__()

        if self.engine == EngineType.TERRAFORM:
            self._output_handler = tf_outputs.TerraformOutputsHandler(
                project_path=self.project_path, project_manifest=self.project_manifest
            )
            self._variable_handler = tf_variables.TerraformVariablesHandler(
                project_path=self.project_path, project_manifest=self.project_manifest
            )
        else:
            raise NotImplementedError(f"OutputsHandler implementation not found for engine: {self.engine}")

    def add_users(self, users: list[str]) -> None:
        """Allowlist the users to access the Jupyter app."""
        command = self.project_manifest.get_command("users.add")
        console = self.get_console()
        runner = cmd_runner.ManifestCommandRunner(
            console=console, output_handler=self._output_handler, variable_handler=self._variable_handler
        )
        joined_users = ",".join(users)

        success, _ = runner.run_command_sequence(
            command,
            cli_paramdefs={
                "users": StrResolvedCliParameter(parameter_name="users", value=joined_users),
                "action": StrResolvedCliParameter(parameter_name="action", value="add"),
                "category": StrResolvedCliParameter(parameter_name="category", value="users"),
            },
        )
        if success:
            runner.update_variables(command)

    def remove_users(self, users: list[str]) -> None:
        """Remove the users from the allowlist of the Jupyter app."""
        command = self.project_manifest.get_command("users.remove")
        console = self.get_console()
        runner = cmd_runner.ManifestCommandRunner(
            console=console, output_handler=self._output_handler, variable_handler=self._variable_handler
        )
        joined_users = ",".join(users)

        success, _ = runner.run_command_sequence(
            command,
            cli_paramdefs={
                "users": StrResolvedCliParameter(parameter_name="users", value=joined_users),
                "action": StrResolvedCliParameter(parameter_name="action", value="remove"),
                "category": StrResolvedCliParameter(parameter_name="category", value="users"),
            },
        )
        if success:
            runner.update_variables(command)

    def set_users(self, users: list[str]) -> None:
        """Replace the list of users allowlisted to access the Jupyter app."""
        command = self.project_manifest.get_command("users.set")
        console = self.get_console()
        runner = cmd_runner.ManifestCommandRunner(
            console=console, output_handler=self._output_handler, variable_handler=self._variable_handler
        )
        joined_users = ",".join(users)

        success, _ = runner.run_command_sequence(
            command,
            cli_paramdefs={
                "users": StrResolvedCliParameter(parameter_name="users", value=joined_users),
                "action": StrResolvedCliParameter(parameter_name="action", value="set"),
                "category": StrResolvedCliParameter(parameter_name="category", value="users"),
            },
        )
        if success:
            runner.update_variables(command)

    def list_users(self) -> list[str]:
        """Return a list of users allowlisted to access the Jupyter app."""
        command = self.project_manifest.get_command("users.list")
        console = self.get_console()
        runner = cmd_runner.ManifestCommandRunner(
            console=console, output_handler=self._output_handler, variable_handler=self._variable_handler
        )
        success, _ = runner.run_command_sequence(
            command,
            cli_paramdefs={
                "category": StrResolvedCliParameter(parameter_name="category", value="users"),
            },
        )
        if success:
            # No variables to update for list command, but we can add the check for consistency
            return runner.get_result_value(command, "users.list", list[str])
        return []



import unittest
from pathlib import Path
from unittest.mock import MagicMock, patch

from jupyter_deploy.engine.enum import EngineType
from jupyter_deploy.handlers.init_handler import InitHandler
from jupyter_deploy.infrastructure.enum import AWSInfrastructureType
from jupyter_deploy.provider.enum import ProviderType


class TestInitHandler(unittest.TestCase):
    """Test class for InitHandler."""

    @patch("jupyter_deploy.fs_utils.get_default_project_path")
    @patch("jupyter_deploy.handlers.init_handler.InitHandler._find_template_path")
    def test_init_with_project_dir(
        self, mock_find_template_path: MagicMock, mock_get_default_project_path: MagicMock
    ) -> None:
        """Test initialization with project_dir provided."""
        # Setup
        project_dir = "/test/project/dir"
        mock_template_path = Path("/mock/template/path")
        mock_find_template_path.return_value = mock_template_path

        # Execute
        handler = InitHandler(project_dir=project_dir)

        # Assert
        self.assertEqual(handler.project_path, Path(project_dir))
        self.assertEqual(handler.engine, EngineType.TERRAFORM)
        mock_find_template_path.assert_called_once_with("aws:ec2:base")
        mock_get_default_project_path.assert_not_called()

    @patch("jupyter_deploy.fs_utils.get_default_project_path")
    @patch("jupyter_deploy.handlers.init_handler.InitHandler._find_template_path")
    def test_init_without_project_dir(
        self, mock_find_template_path: MagicMock, mock_get_default_project_path: MagicMock
    ) -> None:
        """Test initialization without project_dir provided."""
        # Setup
        mock_default_path = Path("/default/project/path")
        mock_template_path = Path("/mock/template/path")
        mock_get_default_project_path.return_value = mock_default_path
        mock_find_template_path.return_value = mock_template_path

        # Execute
        handler = InitHandler(project_dir=None)

        # Assert
        self.assertEqual(handler.project_path, mock_default_path)
        self.assertEqual(handler.engine, EngineType.TERRAFORM)
        mock_find_template_path.assert_called_once_with("aws:ec2:base")
        mock_get_default_project_path.assert_called_once()

    @patch("jupyter_deploy.handlers.init_handler.InitHandler._find_template_path")
    def test_init_with_enum_parameters(self, mock_find_template_path: MagicMock) -> None:
        """Test initialization with enum types for provider and infrastructure."""
        # Setup
        project_dir = "/test/project/dir"
        mock_template_path = Path("/mock/template/path")
        mock_find_template_path.return_value = mock_template_path

        # Execute
        handler = InitHandler(
            project_dir=project_dir,
            engine=EngineType.TERRAFORM,
            provider=ProviderType.AWS,
            infrastructure=AWSInfrastructureType.EC2,
            template="custom-template",
        )

        # Assert
        self.assertEqual(handler.project_path, Path(project_dir))
        self.assertEqual(handler.engine, EngineType.TERRAFORM)
        mock_find_template_path.assert_called_once_with("aws:ec2:custom-template")

    @patch("jupyter_deploy.handlers.init_handler.InitHandler._find_template_path")
    @patch("jupyter_deploy.template_utils.TEMPLATES", {"terraform": {"aws:ec2:base": Path("/mock/template/path")}})
    def test_find_template_path_valid(self, mock_find_template_path: MagicMock) -> None:
        """Test _find_template_path with valid template name."""
        # Setup
        mock_find_template_path.return_value = Path("/mock/template/path")

        handler = InitHandler(project_dir="/test/project/dir")

        # Execute
        mock_find_template_path.return_value = Path("/mock/template/path")
        result = handler._find_template_path("aws:ec2:base")

        # Assert
        self.assertEqual(result, Path("/mock/template/path"))

    @patch("jupyter_deploy.handlers.init_handler.InitHandler._find_template_path")
    def test_find_template_path_empty(self, mock_find_template_path: MagicMock) -> None:
        """Test _find_template_path with empty template name."""
        # Setup
        mock_find_template_path.side_effect = [
            Path("/mock/template/path"),  # For constructor
            ValueError("Template name cannot be empty"),  # For the actual test
        ]

        handler = InitHandler(project_dir="/test/project/dir")

        # Execute and Assert
        with self.assertRaisesRegex(ValueError, "Template name cannot be empty"):
            # Set up the mock to raise the expected exception
            mock_find_template_path.side_effect = ValueError("Template name cannot be empty")
            handler._find_template_path("")

    @patch("jupyter_deploy.handlers.init_handler.InitHandler._find_template_path")
    @patch("jupyter_deploy.template_utils.TEMPLATES", {})
    def test_find_template_path_unsupported_engine(self, mock_find_template_path: MagicMock) -> None:
        """Test _find_template_path with unsupported engine."""
        # Setup
        mock_find_template_path.side_effect = [
            Path("/mock/template/path"),  # For constructor
            ValueError("Engine 'terraform' is not supported. Available engines: none available"),  # For the actual test
        ]

        handler = InitHandler(project_dir="/test/project/dir")

        # Execute and Assert
        with self.assertRaisesRegex(ValueError, "Engine 'terraform' is not supported"):
            mock_find_template_path.side_effect = ValueError("Engine 'terraform' is not supported")
            handler._find_template_path("aws:ec2:base")

    @patch("jupyter_deploy.handlers.init_handler.InitHandler._find_template_path")
    @patch("jupyter_deploy.template_utils.TEMPLATES", {"terraform": {}})
    def test_find_template_path_template_not_found(self, mock_find_template_path: MagicMock) -> None:
        """Test _find_template_path with template not found."""
        # Setup
        mock_find_template_path.side_effect = [
            Path("/mock/template/path"),  # For constructor
            ValueError(
                "Template 'aws:ec2:base' not found for engine 'terraform'. Available templates: none"
            ),  # For the actual test
        ]

        handler = InitHandler(project_dir="/test/project/dir")

        # Execute and Assert
        with self.assertRaisesRegex(ValueError, "Template 'aws:ec2:base' not found"):
            mock_find_template_path.side_effect = ValueError("Template 'aws:ec2:base' not found")
            handler._find_template_path("aws:ec2:base")

    @patch("jupyter_deploy.fs_utils.is_empty_dir")
    @patch("jupyter_deploy.handlers.init_handler.InitHandler._find_template_path")
    def test_may_export_to_project_path_not_exists(
        self, mock_find_template_path: MagicMock, mock_is_empty_dir: MagicMock
    ) -> None:
        """Test may_export_to_project_path when project path doesn't exist."""
        # Setup
        mock_path = MagicMock()
        mock_exists = MagicMock(return_value=False)
        mock_path.exists = mock_exists
        mock_find_template_path.return_value = Path("/mock/template/path")

        handler = InitHandler(project_dir="/test/project/dir")
        handler.project_path = mock_path

        # Execute
        result = handler.may_export_to_project_path()

        # Assert
        self.assertTrue(result)
        mock_exists.assert_called_once()
        mock_is_empty_dir.assert_not_called()

    @patch("jupyter_deploy.fs_utils.is_empty_dir")
    @patch("jupyter_deploy.handlers.init_handler.InitHandler._find_template_path")
    def test_may_export_to_project_path_exists_empty(
        self, mock_find_template_path: MagicMock, mock_is_empty_dir: MagicMock
    ) -> None:
        """Test may_export_to_project_path when project path exists and is empty."""
        # Setup
        mock_path = MagicMock()
        mock_exists = MagicMock(return_value=True)
        mock_path.exists = mock_exists
        mock_is_empty_dir.return_value = True
        mock_find_template_path.return_value = Path("/mock/template/path")

        handler = InitHandler(project_dir="/test/project/dir")
        handler.project_path = mock_path

        # Execute
        result = handler.may_export_to_project_path()

        # Assert
        self.assertTrue(result)
        mock_exists.assert_called_once()
        mock_is_empty_dir.assert_called_once_with(mock_path)

    @patch("jupyter_deploy.fs_utils.is_empty_dir")
    @patch("jupyter_deploy.handlers.init_handler.InitHandler._find_template_path")
    def test_may_export_to_project_path_exists_not_empty(
        self, mock_find_template_path: MagicMock, mock_is_empty_dir: MagicMock
    ) -> None:
        """Test may_export_to_project_path when project path exists and is not empty."""
        # Setup
        mock_path = MagicMock()
        mock_exists = MagicMock(return_value=True)
        mock_path.exists = mock_exists
        mock_is_empty_dir.return_value = False
        mock_find_template_path.return_value = Path("/mock/template/path")

        handler = InitHandler(project_dir="/test/project/dir")
        handler.project_path = mock_path

        # Execute
        result = handler.may_export_to_project_path()

        # Assert
        self.assertFalse(result)
        mock_exists.assert_called_once()
        mock_is_empty_dir.assert_called_once_with(mock_path)

    @patch("jupyter_deploy.fs_utils.safe_clean_directory")
    @patch("jupyter_deploy.handlers.init_handler.InitHandler._find_template_path")
    def test_clear_project_path(self, mock_find_template_path: MagicMock, mock_safe_clean_directory: MagicMock) -> None:
        """Test clear_project_path calls fs_utils.safe_clean_directory with correct path."""
        # Setup
        mock_path = Path("/test/project/dir")
        mock_find_template_path.return_value = Path("/mock/template/path")

        handler = InitHandler(project_dir="/test/project/dir")
        handler.project_path = mock_path

        # Execute
        handler.clear_project_path()

        # Assert
        mock_safe_clean_directory.assert_called_once_with(mock_path)

    @patch("jupyter_deploy.fs_utils.safe_copy_tree")
    @patch("jupyter_deploy.handlers.init_handler.InitHandler._find_template_path")
    def test_setup(self, mock_find_template_path: MagicMock, mock_safe_copy_tree: MagicMock) -> None:
        """Test setup calls fs_utils.safe_copy_tree with correct paths."""
        # Setup
        mock_project_path = Path("/test/project/dir")
        mock_source_path = Path("/mock/template/path")
        mock_find_template_path.return_value = mock_source_path

        handler = InitHandler(project_dir="/test/project/dir")
        handler.project_path = mock_project_path

        # Execute
        handler.setup()

        # Assert
        mock_safe_copy_tree.assert_called_once_with(mock_source_path, mock_project_path)



from collections.abc import Callable
from typing import Any, Generic, TypeVar, get_args

import typer
from pydantic import BaseModel, ConfigDict, ValidationError

from jupyter_deploy import str_utils

T = TypeVar("T")


class TemplateVariableDefinition(BaseModel, Generic[T]):
    """Wrapper class for user-inputable value in a template."""

    model_config = ConfigDict(extra="allow")
    variable_name: str
    description: str
    sensitive: bool = False
    default: T | None = None
    assigned_value: T | None = None
    has_default: bool = False

    def get_cli_var_name(self) -> str:
        """Return variable name using the kebab-case format."""
        cli_var_name = str_utils.to_cli_option_name(self.variable_name)
        return cli_var_name

    def get_cli_description(self) -> str:
        """Return a one-liner description with preset information for the CLI attribute."""
        header = str_utils.get_trimmed_header(self.description)
        default_marker = f"[preset: {self.default}]" if self.default is not None else ""
        separator = " " if len(header) > 0 and len(default_marker) else ""
        return f"{header}{separator}{default_marker}"

    def get_validator_callback(self) -> Callable | None:
        """Return a validator to be passed as callback attribute of the typer.Option()."""
        return None

    def validate_value(self, value: Any) -> T:
        """Verify type to be assigned, return the cast value.

        The value must not be None.

        Raises:
            - ValueError if the value is None
            - TypeError if the value is not of the right type.
        """
        if value is None:
            raise ValueError(f"Attempted to set a None value for variable: {self.variable_name}")

        try:
            dict_val = self.model_dump()
            del dict_val["assigned_value"]
            instance = self.__class__(**dict_val, assigned_value=value)
        except ValidationError as e:
            raise TypeError(f"Invalid value for variable '{self.variable_name}': {value}") from e

        assigned_value = instance.assigned_value

        if assigned_value is None:
            raise ValueError(f"Unexpected assigned value: {self.variable_name}")
        return assigned_value

    @classmethod
    def get_type(cls) -> type:
        """Return the type of the variable."""
        default_field = cls.model_fields["default"]
        type_args = get_args(default_field.annotation)
        return type_args[0]  # type: ignore

    @classmethod
    def convert_assigned_value(cls, assigned_value: Any | None) -> T | None:
        """Convert the value assigned by typer.

        Pass-through with type ignore unless overridden.
        """
        return assigned_value  # type: ignore


class StrTemplateVariableDefinition(TemplateVariableDefinition[str]):
    """Wrapper class for user-inputable string value in a template."""

    pass


class IntTemplateVariableDefinition(TemplateVariableDefinition[int]):
    """Wrapper class for user-inputable integer value in a template."""

    pass


class FloatTemplateVariableDefinition(TemplateVariableDefinition[float]):
    """Wrapper class for user-inputable float value in a template."""

    pass


class AnyNumericTemplateVariableDefinition(TemplateVariableDefinition[int | float]):
    """Wrapper class for user-inputable int or float value in a template."""

    pass


class BoolTemplateVariableDefinition(TemplateVariableDefinition[bool]):
    """Wrapper class for user-inputable bool value in a template."""

    pass


# keep List[str] instead of list[str] here: typer requires it
class ListStrTemplateVariableDefinition(TemplateVariableDefinition[list[str]]):
    """Wrapper class for user-inputable list of string value in a template."""

    def get_cli_description(self) -> str:
        trimmed_description = str_utils.get_trimmed_header(self.description)
        cli_name = self.get_cli_var_name()
        hint = f"Pass each item separately; e.g. --{cli_name} <first-value> --{cli_name} <second-value>"
        return f"{trimmed_description}\n{hint}"

    pass


class ListMapStrTemplateVariableDefinition(TemplateVariableDefinition[list[dict[str, str]]]):
    """Wrapper class for user-inputable list of maps with string keys and values."""

    def get_cli_description(self) -> str:
        trimmed_description = str_utils.get_trimmed_header(self.description)
        cli_name = self.get_cli_var_name()
        hint = (
            "Pass each entry separately, and use comma to separate key=value pairs; "
            f"e.g. --{cli_name} <key1=val1,key2=value2> --{cli_name} <key1=val1,key2=value2>"
        )
        return f"{trimmed_description}\n{hint}"

    def get_validator_callback(self) -> Callable | None:
        def cb(entries: list[str] | None) -> list[str] | None:
            if not entries:
                return []

            for idx, entry in enumerate(entries):
                if not entry:
                    raise typer.BadParameter(f"Empty value at index {idx}")

                key_value_pairs = entry.split(",")
                if not key_value_pairs:
                    raise typer.BadParameter(f"Value at index {idx} must have at least one key=value")
                for key_idx, key_value_pair in enumerate(key_value_pairs):
                    if not isinstance(key_value_pair, str):
                        continue
                    key_value_parts = key_value_pair.split("=")

                    if len(key_value_parts) != 2 or not key_value_parts[0] or not key_value_parts[1]:
                        raise typer.BadParameter(
                            f"Invalid key=value pair for entry {idx} at index {key_idx}, must be of the form key=val"
                        )

                    if not isinstance(key_value_parts[0], str) or not isinstance(key_value_parts[1], str):
                        raise typer.BadParameter(
                            f"Invalid key=value pair for entry {idx} at index {key_idx}, "
                            f"both key and value must be non-empty string"
                        )
            return entries

        return cb

    @classmethod
    def get_type(cls) -> type:
        # this is not a typo!
        # we tell typer that the value is list[str] even though it's a list[dict[str,str]],
        # then we expect user to pass --var-name key1=val1,key2=val2 --var-name key1=val1,key2=val2
        # which we combine afterwards
        return list[str]

    @classmethod
    def convert_assigned_value(cls, assigned_value: Any | None) -> list[dict[str, str]] | None:
        if assigned_value is None:
            return None
        if isinstance(assigned_value, list):
            out: list[dict[str, str]] = []
            for v in assigned_value:  # guaranteed by typer, but keep to mypy happy
                if not isinstance(v, str):
                    continue
                entry: dict[str, str] = {}
                key_value_pairs = v.split(",")

                for key_value_pair in key_value_pairs:
                    key_value_parts = key_value_pair.split("=")
                    entry.update({key_value_parts[0]: key_value_parts[1]})

                out.append(entry)
            return out
        raise ValueError(f"Invalid value: {assigned_value}")


class DictStrTemplateVariableDefinition(TemplateVariableDefinition[dict[str, str]]):
    """Wrapper class for user-inputable dict value whose keys and values are string in a template."""

    def get_cli_description(self) -> str:
        trimmed_description = str_utils.get_trimmed_header(self.description)
        cli_name = self.get_cli_var_name()
        hint = f"Pass the key=value pairs separately; e.g. --{cli_name} <key1=val1> --{cli_name} <key2=val2>"
        return f"{trimmed_description}\n{hint}"

    def get_validator_callback(self) -> Callable | None:
        def cb(entries: list[str] | None) -> list[str] | None:
            if not entries:
                return []

            for idx, v in enumerate(entries):
                if not v:
                    raise typer.BadParameter(f"Empty value at index {idx}, must be of the form key=val")
                parts = v.split("=")
                if len(parts) != 2 or not parts[0] or not parts[1]:
                    raise typer.BadParameter(f"Invalid value at index {idx}, must be of the form key=val")
            return entries

        return cb

    @classmethod
    def get_type(cls) -> type:
        # this is not a typo!
        # we tell typer that the value is list even though it's a dict, then we
        # expect user to pass --var-name key:value --var-name key:value
        # which we combine afterwards
        return list[str]

    @classmethod
    def convert_assigned_value(cls, assigned_value: Any | None) -> dict[str, str] | None:
        if assigned_value is None:
            return None
        if isinstance(assigned_value, list):
            out: dict[str, str] = {}
            for v in assigned_value:  # guaranteed by typer, but keep to mypy happy
                if not isinstance(v, str):
                    continue
                parts = v.split("=")

                if len(parts) != 2:  # guaranteed to be true by validator above
                    continue
                out.update({parts[0]: parts[1]})
            return out

        raise ValueError(f"Invalid value: {assigned_value}")



import unittest
from pathlib import Path
from unittest.mock import Mock, patch

from jupyter_deploy.engine.outdefs import StrTemplateOutputDefinition
from jupyter_deploy.engine.terraform.tf_open import TerraformOpenHandler


class TestTerraformOpenHandler(unittest.TestCase):
    def get_mock_outputs_handler_and_fns(self) -> tuple[Mock, dict[str, Mock]]:
        """Return the mock outputs handler."""
        mock_handler = Mock()
        mock_get_declared_output_def = Mock()
        mock_handler.get_declared_output_def = mock_get_declared_output_def

        mock_get_declared_output_def.return_value = StrTemplateOutputDefinition(
            output_name="notebook_url", value="https://notebook.my.domain"
        )

        return mock_handler, {"get_declared_output_def": mock_get_declared_output_def}

    @patch("jupyter_deploy.engine.terraform.tf_outputs.TerraformOutputsHandler")
    def test_init(self, tf_outputs_handler_cls: Mock) -> None:
        """Test that the TerraformOpenHandler initializes correctly."""
        mock_outputs_handler, outputs_handler_fns = self.get_mock_outputs_handler_and_fns()
        tf_outputs_handler_cls.return_value = mock_outputs_handler
        mock_manifest = Mock()
        handler = TerraformOpenHandler(project_path=Path("/fake/path"), project_manifest=mock_manifest)
        self.assertEqual(handler.project_path, Path("/fake/path"))
        self.assertEqual(handler.project_manifest, mock_manifest)
        self.assertEqual(handler.output_handler, mock_outputs_handler)
        outputs_handler_fns["get_declared_output_def"].assert_not_called()



from typing import Annotated

import typer

from jupyter_deploy import cmd_utils
from jupyter_deploy.handlers.resource import server_handler
from jupyter_deploy.manifest import InvalidServiceError

servers_app = typer.Typer(
    help=("""Interact with the services running your Jupyter app."""),
    no_args_is_help=True,
)


@servers_app.command()
def status(
    project_dir: Annotated[
        str | None,
        typer.Option(
            "--path", "-p", help="Directory of the jupyter-deploy project whose server to send an health check."
        ),
    ] = None,
) -> None:
    """Sends a health check to the services.

    Run either from a jupyter-deploy project directory that you created with `jd init`;
    or pass a --path PATH to such a directory.
    """
    with cmd_utils.project_dir(project_dir):
        handler = server_handler.ServerHandler()
        console = handler.get_console()
        server_status = handler.get_server_status()

        console.print(f"Jupyter server status: [bold cyan]{server_status}[/]")


@servers_app.command()
def start(
    project_dir: Annotated[
        str | None,
        typer.Option("--path", "-p", help="Directory of the jupyter-deploy project whose server to start."),
    ] = None,
    service: Annotated[
        str, typer.Option("--service", "-s", help="Service to start ('all', 'jupyter', or other available services).")
    ] = "all",
) -> None:
    """Start the services.

    By default, starts all services. Specify --service to target a specific service.

    Run either from a jupyter-deploy project directory that you created with `jd init`;
    or pass a --path PATH to such a directory.
    """
    with cmd_utils.project_dir(project_dir):
        handler = server_handler.ServerHandler()
        console = handler.get_console()

        try:
            handler.start_server(service)
        except InvalidServiceError as e:
            console.print(f":x: {e}", style="red")
            return

        if service == "all":
            console.print("Started the Jupyter server and all the sidecars.", style="bold green")
        else:
            console.print(f"Started the '{service}' service.", style="bold green")


@servers_app.command()
def stop(
    project_dir: Annotated[
        str | None,
        typer.Option("--path", "-p", help="Directory of the jupyter-deploy project whose server to stop."),
    ] = None,
    service: Annotated[
        str, typer.Option("--service", "-s", help="Service to stop ('all', 'jupyter', or other available services).")
    ] = "all",
) -> None:
    """Stop the services.

    By default, stops all services. Specify --service to target a specific service.

    Run either from a jupyter-deploy project directory that you created with `jd init`;
    or pass a --path PATH to such a directory.
    """
    with cmd_utils.project_dir(project_dir):
        handler = server_handler.ServerHandler()
        console = handler.get_console()

        try:
            handler.stop_server(service)
        except InvalidServiceError as e:
            console.print(f":x: {e}", style="red")
            return

        if service == "all":
            console.print("Stopped the Jupyter server and all the sidecars.", style="bold green")
        else:
            console.print(f"Stopped the '{service}' service.", style="bold green")


@servers_app.command()
def restart(
    project_dir: Annotated[
        str | None,
        typer.Option("--path", "-p", help="Directory of the jupyter-deploy project whose server to restart."),
    ] = None,
    service: Annotated[
        str, typer.Option("--service", "-s", help="Service to restart ('all', 'jupyter', or other available services).")
    ] = "all",
) -> None:
    """Restart the services.

    By default, restarts all services. Specify --service to target a specific service.

    Run either from a jupyter-deploy project directory that you created with `jd init`;
    or pass a --path PATH to such a directory.
    """
    with cmd_utils.project_dir(project_dir):
        handler = server_handler.ServerHandler()
        console = handler.get_console()

        try:
            handler.restart_server(service)
        except InvalidServiceError as e:
            console.print(f":x: {e}", style="red")
            return

        if service == "all":
            console.print("Restarted the Jupyter server and all the sidecars.", style="bold green")
        else:
            console.print(f"Restarted the '{service}' service.", style="bold green")


@servers_app.command(context_settings={"allow_extra_args": True, "allow_interspersed_args": False})
def logs(
    ctx: typer.Context,
    project_dir: Annotated[
        str | None,
        typer.Option("--path", "-p", help="Directory of the jupyter-deploy project whose server to restart."),
    ] = None,
    service: Annotated[
        str, typer.Option("--service", "-s", help="Name of the service whose logs to display.")
    ] = "default",
) -> None:
    """Print the logs of the service to terminal.

    By default, logs your main application service. Specify --service to target a specific service.

    Run either from a jupyter-deploy project directory that you created with `jd init`;
    or pass a --path <PATH> to such a directory.

    You can pass additional arguments after '--'

    For example, if the underlying engine is docker, use <jd server logs -- -n 100> to retrieve 100 log lines.

    To apply host-side filters, use <jd server logs -- "| grep SEARCH_VALUE">

    Note: invalid characters may prevent logs to be displayed. To view the full logs, connect to your host
    with <jd host connect>.
    """
    # Arguments after -- are in ctx.args
    extra = ctx.args

    with cmd_utils.project_dir(project_dir):
        handler = server_handler.ServerHandler()
        console = handler.get_console()

        try:
            logs, err_logs = handler.get_server_logs(service=service, extra=extra)
        except InvalidServiceError as e:
            console.print(f":x: {e}", style="red")
            return

        if logs:
            console.rule("stdout")
            console.print(logs)
            if not err_logs:
                console.rule()
        else:
            console.print(":warning: no logs were retrieved.", style="yellow")

        if err_logs:
            console.rule("stderr")
            console.print(err_logs)
            console.rule()



from abc import ABC, abstractmethod
from pathlib import Path

from jupyter_deploy.engine import outdefs
from jupyter_deploy.engine.engine_outputs import EngineOutputsHandler
from jupyter_deploy.manifest import JupyterDeployManifest


class EngineDownHandler(ABC):
    def __init__(
        self, project_path: Path, project_manifest: JupyterDeployManifest, output_handler: EngineOutputsHandler
    ) -> None:
        """Instantiate the base handler for `jd down` command."""
        self.project_path = project_path
        self.project_manifest = project_manifest
        self.output_handler = output_handler
        self.engine = project_manifest.get_engine()

    @abstractmethod
    def destroy(self, auto_approve: bool = False) -> None:
        pass

    def get_persisting_resources(self) -> list[str]:
        """Read the manifest value, return the list of resource identifiers.

        Return the empty list if the manifest value is not set.
        """
        try:
            resource_outdef = self.output_handler.get_declared_output_def(
                "persisting_resources", outdefs.ListStrTemplateOutputDefinition
            )
            persisting_resources = resource_outdef.value

            if not persisting_resources:
                return []
            return persisting_resources

        except (ValueError, KeyError, NotImplementedError, TypeError):
            return []
